# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Results:
    def nsort(self, item="", comp="", order="", kabs="", numb="", sel="", **kwargs):
        """Sorts nodal data.

        APDL Command: NSORT

        Parameters
        ----------
        item
            Label identifying the item to be sorted on.  Valid item labels are
            shown in the table below.  Some items also require a component
            label.

        comp
            Component of the item (if required).  Valid component labels are
            shown in the table below.

        order
            Order of sort operation:

            0 - Sort into descending order.

            1 - Sort into ascending order.

        kabs
            Absolute value key:

            0 - Sort according to real value.

            1 - Sort according to absolute value.

        numb
            Number of nodal data records to be sorted in ascending or
            descending order (ORDER) before sort is stopped (remainder will be
            in unsorted sequence) (defaults to all nodes).

        sel
            Allows selection of nodes in the sorted field.

            (blank) - No selection (default).

            SELECT - Select the nodes in the sorted list.

        Notes
        -----
        Values are in the active coordinate system [CSYS for input data or RSYS
        for results data].  Various element results also depend upon the
        recalculation method and the selected results location [AVPRIN, RSYS,
        SHELL, ESEL, and NSEL].  If simultaneous load cases are stored, the
        last sorted sequence formed from any load case applies to all load
        cases.  Use NUSORT to restore the original order.   This command is not
        valid with PowerGraphics.

        Table: 213:: : NSORT - Valid Item and Component Labels

        Table: 214:: : NSORT - Valid Item and Component Labels for Nodal DOF
        Result Values

        Table: 215:: : NSORT - Valid Item and Component Labels for Element
        Result Values

        Works only if failure criteria information is provided. (For more
        information, see the documentation for the FC and TB commands.)

        Must be added via the FCTYP command first.
        """
        command = f"NSORT,{item},{comp},{order},{kabs},{numb},{sel}"
        return self.run(command, **kwargs)

    def nusort(self, **kwargs):
        """Restores original order for nodal data.

        APDL Command: NUSORT

        Notes
        -----
        This command restores the nodal data to its original order (sorted in
        ascending node number sequence) after an NSORT command.  Changing the
        selected nodal set [NSEL] also restores the original nodal order.
        """
        command = f"NUSORT,"
        return self.run(command, **kwargs)

    def plcint(self, action="", id_="", node="", cont="", dtype="", **kwargs):
        """Plots the fracture parameter (CINT) result data.

        APDL Command: PLCINT

        Parameters
        ----------
        action
            PATH

            PATH - Plots CINT quantities according to path number (default).

            FRONT - Plots CINT quantities distribution along the crack front.

        id\\_
            Crack ID number.

        node
            Crack tip node number (default = ALL).

        cont
            Contour number (Default = ALL).

        dtype
            Data type to output:

            JINT - J-integral (default)

            IIN1 - Interaction integral 1

            IIN2 - Interaction integral 2

            IIN3 - Interaction integral 3

            K1 - Mode 1 stress-intensity factor

            K2 - Mode 2 stress-intensity factor

            K3 - Mode 3 stress-intensity factor

            G1 - Mode 1 energy release rate

            G2 - Mode 2 energy release rate

            G3 - Mode 3 energy release rate

            GT - Total energy release rate

            MFTX - Total material force X

            MFTY - Total material force Y

            MFTZ - Total material force Z

            TSTRESS - T-stress

            CEXT - Crack extension

            CSTAR - C*-integral

        Notes
        -----
        The PLCINT command is not available for XFEM-based crack growth
        analyses results processing.
        """
        command = f"PLCINT,{action},{id_},{node},{cont},{dtype}"
        return self.run(command, **kwargs)

    def pldisp(self, kund="", **kwargs):
        """Displays the displaced structure.

        APDL Command: PLDISP

        Parameters
        ----------
        kund
            Undisplaced shape key:

            0 - Display only displaced structure.

            1 - Overlay displaced display with similar undisplaced
                display (appearance is system-dependent).

            2 - Same as 1 except overlay with undisplaced edge display
                (appearance is system-dependent).

        Notes
        -----
        Displays the displaced structure for the selected elements.

        For information on true scale plots, refer to the description of the
        /DSCALE command [/DSCALE,,1.0].
        """
        command = f"PLDISP,{kund}"
        return self.run(command, **kwargs)

    def plesol(self, item="", comp="", kund="", fact="", **kwargs):
        """Displays the solution results as discontinuous element contours.

        APDL Command: PLESOL

        Parameters
        ----------
        item
            Label identifying the item.  Valid item labels are shown in
            Table 219: PLESOL - Valid Item and Component Labels for Element
            Results below.  Some items also require a component label.

        comp
            Component of the item (if required).  Valid component labels are
            shown in Table 219: PLESOL - Valid Item and Component Labels for
            Element Results below.

        kund
            Undisplaced shape key:

            0 - Do not overlay undeformed structure display

            1 - Overlay displaced contour plot with undeformed display (appearance is system-
                dependent)

            2 - Overlay displaced contour plot  with undeformed edge display (appearance is
                system-dependent)

        fact
            Scale factor for 2-D display of contact items (defaults to 1). A
            negative scaling factor may be used to invert the display.

        Notes
        -----
        Displays the solution results as element contours discontinuous across
        element boundaries for the selected elements.  For example, PLESOL,S,X
        displays the X component of stress S (that is, the SX stress
        component).  Various element results depend on the calculation method
        and the selected results location (AVPRIN, RSYS, and ESEL).  Contours
        are determined by linear interpolation within each element, unaffected
        by the surrounding elements (i.e., no nodal averaging is performed).
        The discontinuity between contours of adjacent elements is an
        indication of the gradient across elements.  Component results are
        displayed in the active results coordinate system [RSYS] (default is
        the global Cartesian).  See the ETABLE and PLETAB commands for
        displaying items not available through this command (such as line
        element results).

        For PowerGraphics displays [/GRAPHICS,POWER], results are plotted only
        for the model exterior surface.  The items marked with [1] in Table:
        219:: PLESOL - Valid Item and Component Labels for Element Results are
        not supported by PowerGraphics.

        Table: 219:: : PLESOL - Valid Item and Component Labels for Element
        Results
        """
        command = f"PLESOL,{item},{comp},{kund},{fact}"
        return self.run(command, **kwargs)

    def plnsol(self, item="", comp="", kund="", fact="", fileid="", **kwargs):
        """Displays results as continuous contours.

        APDL Command: PLNSOL

        Parameters
        ----------
        item
            Label identifying the item.  Valid item labels are shown in
            Table 220: PLNSOL - Valid Item and Component Labels below.  Some
            items also require a component label.

        comp
            Component of the item (if required).  Valid component labels are
            shown in  Table 220: PLNSOL - Valid Item and Component Labels
            below.

        kund
            Undisplaced shape key:

            0 - Do not overlay undeformed structure display

            1 - Overlay displaced contour plot with undeformed display (appearance is system-
                dependent)

            2 - Overlay displaced contour plot with undeformed edge display (appearance is
                system-dependent)

        fact
            Scale factor for 2-D display for contact items.  Default value is
            1.  A negative scaling factor may be used to invert the display.

        fileid
            The file index number (obtained via the NLDIAG,NRRE,ON command).
            Valid only for Item = NRRE.

        Notes
        -----
        Displays the solution results as continuous contours across element
        boundaries for the selected nodes and elements. For example, PLNSOL,S,X
        displays the X component of stress S (that is,  the SX stress
        component). Various element results depend upon the recalculation
        method and the selected results location [AVPRIN, RSYS, LAYER, SHELL,
        and NSEL].  Contours are determined by linear interpolation within each
        element from the nodal values, which are averaged at a node whenever
        two or more elements connect to the same node (except for FMAG, which
        is summed at the node).

        For PowerGraphics displays [/GRAPHICS,POWER], results are plotted only
        for the model exterior surface.  The items marked with [2] are not
        supported by PowerGraphics. To plot midside nodes, you must first issue
        /EFACET,2.

        Table: 220:: : PLNSOL - Valid Item and Component Labels
        """
        command = f"PLNSOL,{item},{comp},{kund},{fact},{fileid}"
        return self.run(command, **kwargs)

    def plorb(self, **kwargs):
        """Displays the orbital motion of a rotating structure

        APDL Command: PLORB

        Notes
        -----
        When a structure is rotating and the Coriolis or gyroscopic effect is
        taken into account (CORIOLIS), nodes lying on the rotation axis
        generally exhibit an elliptical orbital motion. The PLORB command
        displays the orbit of each rotating node as well as the deformed shape
        at time t = 0 (the real part of the solution).

        To print the characteristics of the orbital path traversed by each
        node, issue the PRORB command.

        The PLORB command is valid for line elements (such as BEAM188, BEAM189,
        PIPE288, and PIPE289).

        Your model must also involve a rotational velocity (OMEGA or CMOMEGA)
        with Coriolis enabled in a stationary reference frame
        (CORIOLIS,,,,RefFrame = ON).

        A SET command should be issued after PLORB to ensure proper output for
        subsequent postprocessing commands.

        The coordinate system for displaying nodal results must be global
        Cartesian (RSYS,KCN = 0).
        """
        command = f"PLORB,"
        return self.run(command, **kwargs)

    def prenergy(
        self,
        energytype="",
        cname1="",
        cname2="",
        cname3="",
        cname4="",
        cname5="",
        cname6="",
        **kwargs,
    ):
        """Prints the total energies of a model or the energies of the specified

        APDL Command: PRENERGY
        components.

        Parameters
        ----------
        energytype
            Type of energies to be printed:

            ALL - All energies are printed: potential, kinetic, artificial hourglass/drill
                  stiffness, contact stabilization energy, and artificial
                  stabilization energy when applicable. This is the default.

            SENE - Potential energy.

            KENE - Kinetic energy.

        cname1, cname2, cname3,…
            Component names for energies of the components printout.

        Notes
        -----
        The PRENERGY command prints out either the total energies of the entire
        model or the energies of the components depending on the Cname1
        specification.

        Only existing components based on elements (defined with the CM
        command) are supported when component energies are listed.

        This command applies to structural elements only.
        """
        command = f"PRENERGY,{energytype},{cname1},{cname2},{cname3},{cname4},{cname5},{cname6}"
        return self.run(command, **kwargs)

    def prorb(self, **kwargs):
        """Prints the orbital motion characteristics of a rotating structure

        APDL Command: PRORB

        Notes
        -----
        When a structure is rotating and the Coriolis or gyroscopic effect is
        taken into account (CORIOLIS), nodes lying on the rotation axis
        generally exhibit an elliptical orbital motion. The PRORB command
        prints out the orbit characteristics A, B, PSI, PHI, YMAX and ZMAX of
        each rotating node, where

        Angles PSI and PHI are in degrees and within the range of -180 through
        +180. For more information about orbit definition, see Orbits in the
        Advanced Analysis Guide.

        To display the characteristics of the orbital path traversed by each
        node, issue the PLORB command.

        The PRORB command is valid for line elements (such as BEAM188, BEAM189,
        PIPE288, and PIPE289).

        Your model must also involve a rotational velocity (OMEGA or CMOMEGA)
        with Coriolis enabled in a stationary reference frame
        (CORIOLIS,,,,RefFrame = ON).

        A SET command should be issued after PRORB to ensure proper output for
        subsequent postprocessing commands.

        The coordinate system for displaying nodal results must be global
        Cartesian (RSYS,KCN = 0).
        """
        command = f"PRORB,"
        return self.run(command, **kwargs)

    def plvect(
        self,
        item="",
        lab2="",
        lab3="",
        labp="",
        mode="",
        loc="",
        edge="",
        kund="",
        **kwargs,
    ):
        """Displays results as vectors.

        APDL Command: PLVECT

        Parameters
        ----------
        item
            Predefined vector item (from Table 223: PLVECT - Valid Item Labels
            below) or a label identifying the i-component of a user-defined
            vector.

        lab2
            Label identifying the j-component of a user-defined vector. In most
            cases, this value must be blank if Item is selected from
            Table 223: PLVECT - Valid Item Labels. Individual principal
            stresses (Item = S) or principal strains (Item = EPxx) may be
            plotted by specifying the value as 1, 2, or 3.

        lab3
            Label identifying the k-component of a user-defined vector.  Must
            be blank if Item is selected from list below or for 2-D user
            defined vector.

        labp
            Label assigned to resultant vector for display labeling (defaults
            to Item).

        mode
            Vector or raster mode override key:

            (blank) - Use the setting of KEY on the /DEVICE command.

            RAST - Use raster mode for PLVECT displays.

            VECT - Use vector mode for PLVECT displays.

        loc
            Vector location for display of field element results:

            ELEM - Display at element centroid (default).

            NODE - Display at element nodes.

        edge
            Edge display override key:

            (blank) - Use the setting of Key on the /EDGE  command.

            OFF - Deactivate the edge display.

            ON - Activate the edge display.

        kund
            Undisplaced shape key:

            0 - Display vectors on undeformed mesh or geometry.

            1 - Display vectors on deformed mesh or geometry.

        Notes
        -----
        Displays various solution results as vectors (arrows) for the selected
        nodes and/or elements (elements must contain at least three nodes that
        are not colinear).  For example, PLVECT,U displays the displacement
        vector for all selected nodes.  For section displays [/TYPE], the
        vectors are shown only on the section face (i.e., cutting plane).  The
        PLVECT display of principal strains and stresses  (Item = S, EPTO,
        EPEL, EPPL, EPCR, or EPTH) on a "cut" of the model (/TYPE,,1 ,5,7,8, or
        9) is not supported.  The resulting plot displays the vectors on all
        selected elements, not on just the sliced surface.  See the /VSCALE
        command to scale vector lengths. Vector magnitudes may be shown as a
        contour display with the PLNSOL command.  Various results also depend
        upon the recalculation method and the selected results location [LAYER,
        SHELL, and NSEL].

        Items may be selected from a set of recognized vector labels (Item) or
        a vector may be defined from up to three scalar labels
        (Item,Lab2,Lab3).  Scalar labels may be user-defined with the ETABLE
        command.  The vectors appear on an element display as arrows showing
        the relative magnitude of the vector and its direction.  The predefined
        items will be shown either at the node or at the element centroid,
        depending on what item is being displayed and depending on the Loc
        setting.  User defined ETABLE items will be shown at the element
        centroid, regardless of the Loc setting. Stress vectors appear as
        arrows at the element centroid, with the arrowheads pointing away from
        each other for tension and toward each other for compression.

        For PowerGraphics, vector arrow displays are generated in Global
        Cartesian (RSYS = 0). All subsequent displays will revert to your
        original coordinate system.

        When vector mode is active (Mode = VECT), use the Z-buffered display
        type [/TYPE,,6] to maximize speed of PLVECT plots  (other hidden
        display types may make plotting slow). For PowerGraphics
        [/GRAPHICS,POWER], the items marked with [1] are not supported  by
        PowerGraphics.

        It is possible to plot principal stresses (Item = S) or principal
        strains (Item = EPxx) individually. To do so, specify a Lab2 value of
        1, 2, or 3. For example, the following are valid commands:

        Table: 223:: : PLVECT - Valid Item Labels

        Not supported by PowerGraphics
        """
        command = f"PLVECT,{item},{lab2},{lab3},{labp},{mode},{loc},{edge},{kund}"
        return self.run(command, **kwargs)

    def prcint(self, id_="", node="", dtype="", **kwargs):
        """Lists the fracture parameter (CINT) results data.

        APDL Command: PRCINT

        Parameters
        ----------
        id\\_
            Crack ID number.

        node
            Crack tip node number. Default = ALL. Valid only for 3-D analysis.

        dtype
            Data type to output:

            JINT - J-integral

            IIN1 - Interaction integral 1

            IIN2 - Interaction integral 2

            IIN3 - Interaction integral 3

            K1 - Mode 1 stress-intensity factor

            K2 - Mode 2 stress-intensity factor

            K3 - Mode 3 stress-intensity factor

            G1 - Mode 1 energy release rate

            G2 - Mode 2 energy release rate

            G3 - Mode 3 energy release rate

            GT - Total energy release rate

            MFTX - Total material force X

            MFTY - Total material force Y

            MFTZ - Total material force Z

            TSTRESS - T-stress

            CEXT - Crack extension

            CSTAR - C*-integral

            STTMAX - Maximum circumferential stress

            PSMAX - Maximum circumferential stress when

        Notes
        -----
        When a crack tip node is defined, the values associated with the
        specified node are listed.

        Dtype = STTMAX or PSMAX are valid for XFEM-based crack growth analyses
        only.

        In an XFEM-based analysis, issue the command using this syntax:

        PRCINT, ID, , STTMAX (or PSMAX)
        """
        command = f"PRCINT,{id_},{node},{dtype}"
        return self.run(command, **kwargs)

    def presol(self, item="", comp="", **kwargs):
        """Prints the solution results for elements.

        APDL Command: PRESOL

        Parameters
        ----------
        item
            Label identifying the item.  Valid item labels are shown in
            Table 224: PRESOL - Valid Item and Component Labels for Element
            Results below.  Some items also require a component label.

        comp
            Component of the item (if required).  Valid component labels are
            shown in Table 224: PRESOL - Valid Item and Component Labels for
            Element Results below.

        Notes
        -----
        Prints the solution results for the selected elements in the sorted
        sequence.  For example, PRESOL,S prints the stress items SX, SY, SZ,
        SXY, SYZ, and SXZ for the node locations of the element.  Component
        results are in the global Cartesian coordinate directions unless
        transformed (RSYS).

        Shell elements print values at the top, then bottom of the element (or
        layer). If KEYOPT(8) = 2 (for SHELL181, SHELL208, SHELL209, SHELL281,
        or ELBOW290), the results are printed in the order TOP, BOT and then
        MID of each element, (or layer). The MID value will be the actual value
        as written to the results file.

        Items are listed as columns of a table versus element number.  An
        exception occurs for item ELEM which uses an element format (all
        applicable line element results are listed per element) instead of a
        tabular format.

        The FORCE command can be used to define which component of the nodal
        load is to be used (static, damping, inertia, or total).  See the
        ETABLE and PRETAB commands for printing items not available through
        this command (such as line element results).

        For PowerGraphics [/GRAPHICS,POWER], results are listed only for the
        element surface.  The items marked with [1] are not supported  by
        PowerGraphics.

        Table: 224:: : PRESOL - Valid Item and Component Labels for Element
        Results

        Failure criteria for virgin material [1][2].

        Default components: Maximum of all failure criteria defined at current
        location (MAX), maximum strain (EMAX), maximum stress (SMAX), Tsai-Wu
        Strength Index (TWSI), inverse of Tsai-Wu Strength Ratio Index (TWSR).

        Other available components: Hashin Fiber Failure (HFIB), Hashin Matrix
        Failure (HMAT), Puck Fiber Failure (PFIB), Puck Matrix Failure (PMAT),
        LaRc03 Fiber Failure (L3FB),  LaRc03 Matrix Failure (L3MT), LaRc04
        Fiber Failure (L4FB), LaRc04 Matrix Failure (L4MT), and any user-
        defined failure criteria (USR1 through USR9) [4].

        Issue the FCTYP command to activate or remove failure criteria.

        Failure criteria based on the effective stresses in the damaged
        material.

        Components: Maximum of all failure criteria defined at current location
        (MAX), fiber tensile failure (FT), fiber compressive failure (FC),
        matrix tensile failure (MT), and matrix compressive (MC).

        Progressive damage parameters.

        Components: Damage status (STAT, 0 - undamaged, 1 - damaged, 2 -
        complete damage), fiber tensile damage variable (FT), fiber compressive
        damage variable (FC), matrix tensile damage variable (MT), matrix
        compressive damage variable (MC), shear damage variable (S),  energy
        dissipated per unit volume (SED), energy per unit volume due to viscous
        damping (SEDV).
        """
        command = f"PRESOL,{item},{comp}"
        return self.run(command, **kwargs)

    def prjsol(self, item="", comp="", **kwargs):
        """Prints joint element output.

        APDL Command: PRJSOL

        Parameters
        ----------
        item
            Label identifying the item. Some items also require a component
            label.

            DISP - Relative displacements.

            ROT - Relative rotations.

            VEL - Relative linear velocities.

            OMG - Relative angular velocities.

            ACC - Relative linear accelerations.

            DMG - Relative angular accelerations.

            SMISC - Summable miscellaneous quantities.

        comp
            Component of the item (if required). For Item = DISP, ROT, VEL,
            OMG, ACC, and DMG, enter the direction label, X, Y, or Z. For Item
            = SMISC, enter a valid number.

        Notes
        -----
        Prints element output for the MPC184 joint element. The joint element
        quantities printed are the values for the free or unconstrained
        relative degrees of freedom.

        This command is valid in POST1 only.
        """
        command = f"PRJSOL,{item},{comp}"
        return self.run(command, **kwargs)

    def prnld(self, lab="", tol="", item="", **kwargs):
        """Prints the summed element nodal loads.

        APDL Command: PRNLD

        Parameters
        ----------
        lab
            Nodal reaction load type.  If blank, use the first ten of all
            available labels.  Valid labels are:

        tol
            Tolerance value about zero within which loads are not printed, as
            follows:

            > 0  - Relative tolerance about zero within which loads are not printed. In this case,
                   the tolerance is TOL * Load, where Load is the absolute
                   value of the maximum load on the selected nodes.

             0  -  Print all nodal loads.

            > 0  - Absolute tolerance about zero within which loads are not printed.

        item
            Selected set of nodes.

            (blank) - Prints the summed element nodal loads for all selected nodes (default),
                      excluding contact elements.

            CONT - Prints the summed element nodal loads for contact nodes only.

            BOTH - Prints the summed element nodal loads for all selected nodes, including contact
                   nodes.

        Notes
        -----
        Prints the summed element nodal loads (forces, moments, heat flows,
        flux, etc.) for the selected nodes in the sorted sequence.  Results are
        in the global Cartesian coordinate directions unless transformed
        [RSYS].  Zero values (within a tolerance range) are not printed.  Loads
        applied to a constrained degree of freedom are not included.  The FORCE
        command can be used to define which component of the nodal load is to
        be used (static, damping, inertia, or total).

        By default, PRNLD excludes elements TARGE169 - CONTA177. Setting ITEM =
        CONT will only account for nodal forces on selected contact elements
        (CONTA171 - CONTA177). Setting ITEM = BOTH will account for nodal
        forces on all selected nodes, including contact nodes.
        """
        command = f"PRNLD,{lab},{tol},{item}"
        return self.run(command, **kwargs)

    def prnsol(self, item="", comp="", **kwargs):
        """Prints nodal solution results.

        APDL Command: PRNSOL

        Parameters
        ----------
        item
            Label identifying the item.  Valid item labels are shown in
            Table 225: PRNSOL - Valid Item and Component Labels below.  Some
            items also require a component label.

        comp
            Component of the item (if required).  Valid component labels are
            shown in Table 225: PRNSOL - Valid Item and Component Labels below.
            Defaults to COMP.

        Notes
        -----
        Prints the nodal solution results for the selected nodes in the sorted
        sequence. For example, PRNSOL,U,X prints the X component of
        displacement vector U (that is, the UX degree of freedom). Component
        results are in the global Cartesian coordinate directions unless
        transformed (RSYS). Various element results also depend upon the
        recalculation method and the selected results location (AVPRIN, RSYS,
        LAYER, SHELL, and NSEL). If the LAYER command is issued, then the
        resulting output is listed in full graphics mode (/GRAPHICS,FULL). You
        can use the FORCE command to define which component of the nodal load
        (static, damping, inertia, or total) should be used.

        PowerGraphics can affect your nodal solution listings. For
        PowerGraphics (/GRAPHICS,POWER), results are listed only for the model
        exterior surfaces.

        When shell element types are present, results are output on a surface-
        by-surface basis. For shell elements, such as SHELL181 or SHELL281, and
        for ELBOW290, printed output is for both the top and bottom surfaces.
        For solid elements such as SOLID185, the output is averaged for each
        surface and printed as follows:

        For a node at a vertex, three lines are output (one printed line for
        each surface).

        For a node on an edge, two lines are output (one printed line for each
        surface).

        For nodes on a face, one value is output.

        For nodes interior to the volume, no printed values are output.

        If a node is common to more than one element, or if a geometric
        discontinuity exists, several conflicting listings may result. For
        example, a corner node incorporating results from solid elements and
        shell elements could yield as many as nine different results; the
        printed output would be averages at the top and bottom for the three
        shell surfaces plus averages at the three surfaces for the solid, for a
        total of nine lines of output. ANSYS does not average result listings
        across geometric discontinuities when shell element types are present.
        It is important to analyze the listings at discontinuities to ascertain
        the significance of each set of data.

        The printed output for full graphics (/GRAPHICS,FULL) follows the
        standard ANSYS convention of averaging results at the node.  For shell
        elements, the default for display is TOP so that the results for the
        top of the shell are averaged with the other elements attached to that
        node.

        If an NSORT, ESORT or /ESHAPE command is issued with PowerGraphics
        activated, then the PRNSOL listings will be the same as in full
        graphics mode (/GRAPHICS,FULL). The items marked with [2] are not
        supported by PowerGraphics. To print midside nodes, you must first
        issue an /EFACET,2 command.

        Table: 225:: : PRNSOL - Valid Item and Component Labels

        Failure criteria  [2][4].

        Default components: Maximum of all failure criteria defined at current
        location (MAX), maximum strain (EMAX), maximum stress (SMAX), Tsai-Wu
        Strength Index (TWSI), inverse of Tsai-Wu Strength Ratio Index (TWSR).
        """
        command = f"PRNSOL,{item},{comp}"
        return self.run(command, **kwargs)

    def prrfor(self, lab="", **kwargs):
        """Prints the constrained node reaction solution. Used with the FORCE

        APDL Command: PRRFOR
        command.

        Parameters
        ----------
        lab
            Nodal reaction load type.  If blank, use the first ten of all
            available labels. Valid labels are:

        Notes
        -----
        PRRFOR has the same functionality as the PRRSOL command; use PRRFOR
        instead of PRRSOL when a FORCE command has been issued.

        In a non-spectrum analysis, if either contact or pretension elements
        exist in the model, PRRFOR uses the PRRSOL command internally and the
        FORCE setting is ignored.

        Because modal displacements cannot be used to calculate contact element
        nodal forces,: those forces are not included in the spectrum and PSD
        analyses reaction solution. As a consequence, the: PRRFOR: command is
        not supported when constraints on contact element pilot nodes are
        present.
        """
        command = f"PRRFOR,{lab}"
        return self.run(command, **kwargs)

    def prrsol(self, lab="", **kwargs):
        """Prints the constrained node reaction solution.

        APDL Command: PRRSOL

        Parameters
        ----------
        lab
            Nodal reaction load type.  If blank, use the first ten of all
            available labels. Valid labels are:

        Notes
        -----
        Prints the constrained node reaction solution for the selected nodes in
        the sorted sequence.  For coupled nodes and nodes in constraint
        equations, the sum of all reactions in the coupled or constraint
        equation set  appears at the primary node of the set.  Results are in
        the global Cartesian coordinate directions unless transformed (RSYS).

        PRRSOL is not valid if any load is applied to a constrained node in the
        direction of the constraint and any of the following is true:

        LCOPER has been used.

        LCASE has been used to read from a load case file.

        The applied loads and constraints in the database are not the ones used
        to create the results data being processed.

        PRRSOL provides the total reaction solution (static, plus damping, plus
        inertial, as appropriate based on the analysis type); however, modal
        reactions include only the static contribution.

        Use PRRFOR instead of PRRSOL with the FORCE command to obtain only the
        static, damping, or inertial components.
        """
        command = f"PRRSOL,{lab}"
        return self.run(command, **kwargs)

    def prvect(self, item="", lab2="", lab3="", labp="", **kwargs):
        """Prints results as vector magnitude and direction cosines.

        APDL Command: PRVECT

        Parameters
        ----------
        item
            Predefined vector item (from Table 226: PRVECT - Valid Item and
            Component Labels below) or a label identifying the i-component of a
            user-defined vector.

        lab2
            Label identifying the j-component of a user-defined vector. In most
            cases, this value must be blank if Item is selected from
            Table 226: PRVECT - Valid Item and Component Labels. Individual
            principal stresses (Item = S) or principal strains (Item = EPxx)
            may be printed by specifying the value as 1, 2, or 3.

        lab3
            Label identifying the k-component of a user-defined vector.  Must
            be blank if Item is selected from list below or for 2-D user
            defined vector.

        labp
            Label assigned to resultant vector for printout labeling (defaults
            to Item).

        Notes
        -----
        Prints various solution results as vector magnitude and direction
        cosines for the selected nodes and/or elements.  For example, PRVECT,U
        prints the displacement magnitude and its direction cosines for all
        selected nodes.  For nodal degree of freedom vector results, direction
        cosines are with respect to the results coordinate system RSYS. For
        element results, direction cosines are with respect to the global
        Cartesian system. Item components may be printed with the PRNSOL
        command.  Various results also depend upon the recalculation method and
        the selected results location [LAYER, SHELL, NSEL, and ESEL].  Items
        may be selected from a set of recognized vector labels (Item) or a
        vector may be defined from up to three scalar labels (Item,Lab2,Lab3).
        Scalar labels may be user-defined with the ETABLE command.

        Portions of this command are not supported by PowerGraphics
        [/GRAPHICS,POWER].

        Table: 226:: : PRVECT - Valid Item and Component Labels
        """
        command = f"PRVECT,{item},{lab2},{lab3},{labp}"
        return self.run(command, **kwargs)

    def sumtype(self, label="", **kwargs):
        """Sets the type of summation to be used in the following load case

        APDL Command: SUMTYPE
        operations.

        Parameters
        ----------
        label
            Summation type

            COMP - Combine element component stresses only.  Stresses such as average nodal
                   stresses, principal stresses, equivalent stresses, and
                   stress intensities are derived from the combined element
                   component stresses. Default.

            PRIN - Combine principal stress, equivalent stress, and stress intensity directly as
                   stored on the results file.  Component stresses are not
                   available with this option.

        Notes
        -----
        Issue SUMTYPE,PRIN when you want to have a load case operation (LCOPER)
        act on the principal / equivalent stresses instead of the component
        stresses. Also issue SUMTYPE,PRIN when you want to read in load cases
        (LCASE). Note that the SUMTYPE setting is not maintained between /POST1
        sessions.

        SUMTYPE,PRIN also causes principal nodal values to be the average of
        the contributing principal element nodal values (see AVPRIN,1).

        BEAM188 and BEAM189 elements compute principal stress, equivalent
        stress, and stress intensity values on request instead of storing them
        on the results file;  SUMTYPE,PRIN does not apply for these elements.
        """
        command = f"SUMTYPE,{label}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Setup:
    def rmresume(self, fname="", ext="", **kwargs):
        """Resumes ROM data from a file.

        APDL Command: RMRESUME

        Parameters
        ----------
        fname
            Name and directory path of the ROM database file (248 character
            maximum). Default to Jobname.

        ext
            Extension of the ROM database file. Default to .rom.

        Notes
        -----
        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMRESUME,{fname},{ext}"
        return self.run(command, **kwargs)

    def rmsave(self, fname="", ext="", **kwargs):
        """Saves ROM data to file.

        APDL Command: RMSAVE

        Parameters
        ----------
        fname
            Name and directory path of the ROM database file. Default to
            Jobname.

        ext
            Extension of the ROM database file. Default to .rom.

        Notes
        -----
        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        return self.run(f"RMSAVE,{fname},{ext}", **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Conn:
    def cat5in(
        self,
        name="",
        extension="",
        path="",
        entity="",
        fmt="",
        nocl="",
        noan="",
        **kwargs,
    ):
        """Transfers a .CATPart file into the ANSYS program.

        APDL Command: ~CAT5IN

        Parameters
        ----------
        name
            The name of a valid .CATPart file, created with CATIA Version 5.0.
            The first character of the file name must be an alphanumeric.

        extension
            The extension for the file. The default extension is .CATPart.

        path
            The path name of the directory in which the file resides enclosed
            in single quotes. The default path name is the current working
            directory.

        entity
            Entity to be imported.

            SOLIDS
                Solids only, imported as ANSYS volumes (default).

            SURFACES
                Surfaces only, imported as ANSYS areas.

            ALL
                All entities. Use this option when the file contains different types of
                entities.

        fmt
            The format in which ANSYS will store the model.

            0
                Neutral format (default). Defeaturing after import is restricted.

            1
                Solid format; this allows defeaturing after import.

        nocl
            Remove tiny objects.

            0
                Remove tiny objects without checking model validity (default).

            1
                Do not remove tiny objects.

        noan
            Perform an analysis of the model.

            0
                Analyze the model (default).

            1
                Do not analyze the model.

        Notes
        -----
        If defeaturing is specified (FMT = 1), this command must be the last
        line of any file, script, or other interactive input.

        More information on importing CATIA Version 5 parts is available in
        CATIA V5 in the Connection User's Guide.
        """
        command = f"~CAT5IN,{name},{extension},{path},{entity},{fmt},{nocl},{noan}"
        return self.run(command, **kwargs)

    def catiain(self, name="", extension="", path="", blank="", **kwargs):
        """Transfers a CATIA model into the ANSYS program.

        APDL Command: ~CATIAIN

        Parameters
        ----------
        name
            The name of a valid CATIA model, created with CATIA 4.x or
            lower.  The first character of the file name must be an
            alphanumeric.  Special characters such as ``&`` ``-`` and ``*`` and
            spaces are not permitted in the part name.

        extension
            The extension for the file. The default extension is .model.

        path
            The path name of the directory in which the file resides, enclosed
            in single quotes. The default path name is the current working
            directory.

        blank
            Sets whether to import "blanked" entities.

            0
                Does not import "blanked" (suppressed) CATIA entities (default).

            1
                Imports "blanked" entities. The portions of CATIA data
                that were suppressed will be included in the import.

        Notes
        -----
        More information on importing CATIA parts is available in CATIA V4 in
        the Connection User's Guide.
        """
        return self.run(f"~CATIAIN,{name},{extension},{path},,,{blank}", **kwargs)

    def parain(
        self,
        name="",
        extension="",
        path="",
        entity="",
        fmt="",
        scale="",
        **kwargs,
    ):
        """Transfers a Parasolid file into the ANSYS program.

        APDL Command: ~PARAIN

        Parameters
        ----------
        name
            The name of a valid Parasolid file. The first character of the file
            name must be an alphanumeric.

        extension
            The extension for the file. The default extension is .x_t on a PC
            or .xmt_txt on a Linux system. Parasolid files are compatible
            across systems, and do not need to be renamed to be used on another
            platform.

        path
            The path name of the directory in which the file resides, enclosed
            in single quotes. The default path name is the current working
            directory.

        entity
            Entity to be imported:

            SOLIDS
                Solids only, imported as ANSYS volumes (default)

            SURFACES
                Surfaces only, imported as ANSYS areas.

            WIREFRAME
                Wireframe only, imported as ANSYS lines.

            ALL
                All entities. Use this option when the file contains
                more than one type of entity.

        fmt
            Sets the format in which ANSYS will store the model

            0
                Neutral format (default). Defeaturing after import is
                restricted. Use this option if you need to scale a model
                to a specific unit of measure (other than meters).

            1
                Solid format; this allows defeaturing after import.

        scale
            Allows scaling for the model

            0
                Do not rescale the model; retain the default Parasolid
                setting of meters (default).

            1
                Scale the model if warranted by the model size.

        Notes
        -----
        More information on importing Parasolid parts is available in Parasolid
        in the Connection User's Guide.
        """
        command = f"~PARAIN,{name},{extension},{path},{entity},{fmt},{scale}"
        return self.run(command, **kwargs)

    def proein(self, name="", extension="", path="", proecomm="", **kwargs):
        """APDL Command: ~PROEIN

        Transfers a Creo Parametric part into the ANSYS program.

        Parameters
        ----------
        name
            The name of the Creo Parametric part to be imported, which cannot
            exceed 64 characters in length and must begin with an alphanumeric
            character. Special characters such as ``&`` ``-`` and ``*`` and spaces are not
            permitted in the part name.

        extension
            The general Creo Parametric extension format is prt for parts and
            asm for assemblies.

        path
            Full path name to the directory containing the part. The default is
            the current working directory.

        proecomm
            The start command for the version of Creo Parametric you are using.
            proe1 is the default command. Note that the full path name to the
            Creo Parametric command need not be used here if the path had been
            included in the PATH variable. The Creo Parametric command name is
            set by the PROE_START_CMD162 environment variable.

        Notes
        -----
        More information on importing Creo Parametric parts is available in
        Creo Parametric (formerly Pro/ENGINEER) in the Connection User's Guide.
        """
        command = f"~PROEIN,{name},{extension},{path},{proecomm}"
        return self.run(command, **kwargs)

    def satin(
        self,
        name="",
        extension="",
        path="",
        entity="",
        fmt="",
        nocl="",
        noan="",
        **kwargs,
    ):
        """Transfers a .SAT file into the ANSYS program.

        APDL Command: ~SATIN

        Parameters
        ----------
        name
            The name of a valid .SAT file, created with a supported version of
            ACIS. The first character of the file name must be an alphanumeric.
            Special characters such as ``&`` ``-`` and ``*`` and spaces are not permitted
            in the part name. See File Names in the Command Reference for more
            information about ANSYS file naming conventions.

        extension
            The extension for the file. The default extension is .sat.

        path
            The path name of the directory in which the file resides enclosed
            in single quotes. The default path name is the current working
            directory.

        entity
            Entity to be imported.

            SOLIDS
                Solids only, imported as ANSYS volumes (Not implemented, imports All).

            SURFACES
                Surfaces only, imported as ANSYS areas (Not implemented, imports All).

            WIREFRAME
                Wireframe only, imported as ANSYS lines (Not implemented, imports All).

            ALL
                All entities. Use this option when the file contains different types of
                entities.

        fmt
            The format in which ANSYS will store the model.

            0
                Neutral format (default). Defeaturing after import is restricted.

            1
                Solid format; this allows defeaturing after import.

        nocl
            Remove tiny objects.

            0
                Remove tiny objects without checking model validity (default).

            1
                Do not remove tiny objects.

        noan
            Perform an ACIS analysis of the model.

            0
                Analyze the model (default).

            1
                Do not analyze the model.

        Notes
        -----
        More information on importing ACIS parts is available in ACIS in the
        Connection User's Guide.
        """
        command = f"~SATIN,{name},{extension},{path},{entity},{fmt},{nocl},{noan}"
        return self.run(command, **kwargs)

    def ugin(
        self,
        name="",
        extension="",
        path="",
        entity="",
        layer="",
        fmt="",
        **kwargs,
    ):
        """Transfers an NX part into the ANSYS program.

        APDL Command: ~UGIN

        Parameters
        ----------
        name
            The file name of the NX part to be imported, which cannot exceed 64
            characters in length. The path name must begin with an alphanumeric
            character. Special characters such as &, -,  and * are not
            permitted in the part name.

        extension
            The NX part file extension. The default is .prt.

        path
            The full path name to the directory containing the part, enclosed
            in single quotes; for example, '/ug_parts'. The default is the
            current working directory.

        entity
            Entity to be imported.

            0 or Solid
                Solids only, imported as ANSYS volumes (the
                default).

            1 or Surface
                Surfaces only, imported as ANSYS areas.

            2 or Wireframe
                Wireframe only, imported as ANSYS lines.

            3 or All
                All entities. Use this option when the part
                contains entities that may not be attached to each
                other, such as a solid in one location and a
                surface in another.

        layer
            The number(s) assigned to the layer(s) to be imported. You can
            import one layer or a range of layers (designated by hyphens).
            Defaults to 1-256 (all layers).

        fmt
            The format in which ANSYS will store the model.

            0
                Neutral format (default). Defeaturing after import is restricted.

            1
                Solid format; this allows defeaturing after import.

        Notes
        -----
        More information on importing NX parts is available in UG/NX in the
        Connection User's Guide.
        """
        command = f"~UGIN,{name},{extension},{path},{entity},{layer},{fmt}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Contains the MapdlDb classes, allowing the access to MAPDL DB from Python."""
from enum import Enum
from functools import wraps
import os
import time
from typing import Optional
from warnings import warn
import weakref

from ansys.api.mapdl.v0 import mapdl_db_pb2_grpc
from ansys.tools.versioning import server_meets_version
import grpc

from ansys.mapdl.core.errors import MapdlConnectionError
from ansys.mapdl.core.mapdl_grpc import MAX_MESSAGE_LENGTH, MapdlGrpc

MINIMUM_MAPDL_VERSION = "21.1"
FAILING_DATABASE_MAPDL = ["24.1", "24.2"]

DEFAULT_DB_PORT = 50055


def check_mapdl_db_is_alive(function):
    """
    Decorator to check that the MAPDL.DB has started.

    It works for the DB object (DBDef) and for the derived object which has "_db" attribute.
    """

    @wraps(function)
    def wrapper(self, *args, **kwargs):
        if hasattr(self, "active"):
            active = self.active
        elif hasattr(self, "_db"):
            active = self._db.active
        else:  # pragma: no cover
            raise Exception("The DB object could not be found.")

        if not active:
            self._mapdl._log.error(
                f"Please start the MAPDL DB Server to access '{function.__name__}'."
            )
            return None
        return function(self, *args, **kwargs)

    return wrapper


class DBDef(Enum):  # From MAPDL ansysdef.inc include file
    """Database type definitions."""

    DB_SELECTED = 1
    DB_NUMDEFINED = 12
    DB_NUMSELECTED = 13
    DB_MAXDEFINED = 14
    DB_MAXRECLENG = 15
    DB_GETNEXTRECD = 16
    DB_MAXALLOC = 17
    DB_OBJFLAG = -777
    DB_NEXT = -888
    DB_NEXT_ALLOC = -889
    DB_MODIFIED = -999
    DB_INTERNAL = -9999


class MapdlDb:
    """
    Abstract mapdl database class.  Created from a ``Mapdl`` instance.

    Examples
    --------
    Create a nodes instance.

    >>> from ansys.mapdl.core import launch_mapdl
    >>> mapdl = launch_mapdl()
    >>> # create nodes...
    >>> nodes = mapdl.db.nodes
    >>> print(nodes)
    MAPDL Database Nodes
        Number of nodes:          270641
        Number of selected nodes: 270641
        Maximum node number:      270641

    >>> mapdl.nsel("NONE")
    >>> print(nodes)
    MAPDL Database Nodes
        Number of nodes:          270641
        Number of selected nodes: 0
        Maximum node number:      270641

    Return the selection status and the coordinates of node 22.

    >>> nodes = mapdl.db.nodes
    >>> sel, coord = nodes.coord(22)
    >>> coord
    (1.0, 0.5, 0.0, 0.0, 0.0, 0.0)

    """

    def __init__(self, mapdl):
        """Initialize this class."""
        if not isinstance(mapdl, MapdlGrpc):  # pragma: no cover
            raise TypeError("``mapdl`` must be a MapdlGrpc instance")
        self._mapdl_weakref = weakref.ref(mapdl)
        self._stub = None
        self._channel = None
        self._ip = None
        self._server = {}
        self._channel_str = None
        self._state = None
        self._nodes = None
        self._elems = None

    def __str__(self):
        """Return the string representation of this class."""
        if self.active:
            return f"MAPDL database server active at {self._channel_str}"
        return f"MAPDL database server not active"

    @property
    def _mapdl(self):
        """Return the weakly referenced instance of mapdl."""
        return self._mapdl_weakref()

    def _start(self, port: Optional[int] = None) -> int:
        """
        Lower level start of the database server.

        Returns
        -------
        int
            Port of the database server.

        """
        # database server must be run from the "BEGIN" level
        self._mapdl._log.debug("Starting MAPDL server")
        self._mapdl._cache_routine()
        with self._mapdl.run_as_routine("Begin level"):
            self._mapdl.run(f"/DBS,SERVER,START,{port}")

        if not port:
            # We do not know which port has started with
            # Scan the DBServer.info file to get the Port Number

            self._mapdl._log.debug("Downloading 'DBServer' file.")
            tstart = time.time()
            timeout = 1
            status = self._mapdl._download_as_raw("DBServer.info").decode()
            while status == "":  # pragma: no cover
                status = self._mapdl._download_as_raw("DBServer.info").decode()
                time.sleep(0.05)
                if time.time() - tstart > timeout:
                    raise TimeoutError(
                        f"Unable to start database server in {timeout} second(s). DBServer not received."
                    )

            self._mapdl._log.debug("Downloading 'DBServer' file.")
            try:
                # expected of the form 'Port : 50055'
                port = int(status.split(":")[1])
            except Exception as e:  # pragma: no cover
                self._mapdl._log.error(
                    "Unable to read port number from '%s' due to\n%s",
                    status,
                    str(e),
                )
                port = DEFAULT_DB_PORT

        self._mapdl._log.debug("MAPDL database server started on port %d", port)
        return port

    @property
    def active(self) -> bool:
        """Return if the database server is active."""
        return "NOT" not in self._status()

    def start(self, port: Optional[int] = None, timeout: int = 10):
        """
        Start the gRPC MAPDL database server.

        Parameters
        ----------
        timeout : float, optional
            Timeout to start the service.

        Examples
        --------
        >>> mapdl.db.start()
        """
        # checking MAPDL API
        from ansys.api.mapdl import __version__ as api_version

        api_version = tuple(int(each) for each in api_version.split("."))

        if api_version < (0, 5, 1):  # pragma: no cover
            raise ImportError(
                "Please upgrade the 'ansys.api.mapdl' package to at least v0.5.1."
                "You can use 'pip install ansys-api-mapdl --upgrade"
            )

        ## Checking MAPDL versions
        mapdl_version = str(self._mapdl.version)
        if (
            not server_meets_version(mapdl_version, MINIMUM_MAPDL_VERSION)
            or mapdl_version in FAILING_DATABASE_MAPDL
        ):  # pragma: no cover
            from ansys.mapdl.core.errors import MapdlVersionError

            raise MapdlVersionError(
                f"This MAPDL version ({mapdl_version}) is not compatible with the Database module."
                "Please check the online documentation regarding Database Module at 'mapdl.docs.pyansys.com'."
            )

        if self._mapdl._server_version < (0, 4, 1):
            from ansys.mapdl.core.errors import MapdlVersionError

            ver_ = ".".join([str(each) for each in self._mapdl._server_version])
            raise MapdlVersionError(
                f"This version of MAPDL gRPC API version ('ansys.api.mapdl' == {ver_}) is not compatible with 'database' module.\n"
                "Please check the online documentation at 'mapdl.docs.pyansys.com' "
            )

        # only start if not already running
        is_running = self.active
        self._mapdl._log.debug("MAPDL DB server running: %s", str(is_running))
        if is_running:
            return

        # permit overriding db_port via env var for CI
        if not port:
            if (
                "PYMAPDL_DB_PORT" in os.environ
                and os.environ.get("PYMAPDL_DB_PORT").isdigit()
            ):
                db_port_str = int(os.environ.get("PYMAPDL_DB_PORT"))
                self._mapdl._log.debug(
                    f"Setting DB port from 'PYMAPDL_DB_PORT' env var: {db_port_str}"
                )
            else:
                self._mapdl._log.debug(
                    f"Setting default DB port ('{DEFAULT_DB_PORT}') because no port was input or the env var 'PYMAPDL_DB_PORT' is not correctly set."
                )
                port = DEFAULT_DB_PORT

        db_port = self._start(port=port)

        if not self._ip:
            self._ip = self._mapdl.ip

        self._server = {"ip": self._ip, "port": db_port}
        self._channel_str = f"{self._ip}:{db_port}"

        self._channel = grpc.insecure_channel(
            self._channel_str,
            options=[
                ("grpc.max_receive_message_length", MAX_MESSAGE_LENGTH),
            ],
        )
        self._state = grpc.channel_ready_future(self._channel)
        self._stub = mapdl_db_pb2_grpc.MapdlDbServiceStub(self._channel)

        # wait until the channel matures
        tstart = time.time()
        while ((time.time() - tstart) < timeout) and not self._state._matured:
            time.sleep(0.01)

        if not self._state._matured:  # pragma: no cover
            raise MapdlConnectionError(
                f"Unable to establish connection to MAPDL database server {self._channel_str}"
            )
        self._mapdl._log.debug("Established connection to MAPDL database server")

    def _stop(self):
        """Stop the MAPDL database service."""
        with self._mapdl.run_as_routine("Begin level"):
            return self._mapdl.run("/DBS,SERVER,STOP")

    def stop(self):
        """
        Shutdown the MAPDL database service and close the connection.

        Examples
        --------
        Stop the database service.

        >>> mapdl.db.stop()
        """
        if not self.active:
            warn("Server is already shutdown")
            return

        self._mapdl._log.debug("Closing the connection with the MAPDL DB Server")
        self._stop()
        self._channel.close()
        self._channel = None
        self._stub = None
        self._state = None

    def _status(self):
        """
        Return the status of the MADPL DB Server.

        Examples
        --------
        >>> output = mapdl.db._status()
        >>> print(output)
        >ENTERING THE SERVER MODE: STATUS

         DB Server is NOT currently running ..
        """
        # Need to use the health check here
        with self._mapdl.run_as_routine("Begin level"):
            return self._mapdl.run("/DBS,SERVER,STATUS")

    def load(self, fname, progress_bar=False):
        """
        Load a MAPDL database file in memory.

        Parameters
        ----------
        fname : str
            The file name of the database to load.

        Examples
        --------
        >>> mapdl.db.load('file.db')
        """
        self._mapdl._get_file_path(fname, progress_bar=progress_bar)
        return self._mapdl.resume(fname)

    def save(self, fname, option="ALL"):
        """
        Save the MAPDL database to disk.

        Parameters
        ----------
        fname : str
            Filename to save the database to.

        option : str
            The mode for saving the database, either:

            * "ALL" - Both the model and the solution
            * "MODEL" - Just the model
            * "SOLU" - Just the solution

        Examples
        --------
        >>> mapdl.db.save('model.db', option=)
        """
        allowed = ["ALL", "MODEL", "SOLU"]
        if option.upper() not in allowed:
            raise ValueError(f"Option must be one of the following: {allowed}")

        return self._mapdl.run(f"SAVE,{fname},,,{option}")

    def clear(self, **kwargs):
        """
        Delete everything in the MAPDL database.

        Examples
        --------
        >>> mapdl.db.clear()
        """
        return self._mapdl.run("/CLEAR,ALL")

    @property
    @check_mapdl_db_is_alive
    def nodes(self):
        """
        MAPDL database nodes interface.

        Returns
        -------
        :class:`DbNodes <ansys.mapdl.core.database.nodes.DbNodes>`

        Examples
        --------
        Create a nodes instance.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> # create nodes...
        >>> nodes = mapdl.db.nodes
        >>> print(nodes)
        MAPDL Database Nodes
            Number of nodes:          270641
            Number of selected nodes: 270641
            Maximum node number:      270641

        >>> mapdl.nsel("NONE")
        >>> print(nodes)
        MAPDL Database Nodes
            Number of nodes:          270641
            Number of selected nodes: 0
            Maximum node number:      270641

        Return the selection status and the coordinates of node 22.

        >>> nodes = mapdl.db.nodes
        >>> sel, coord = nodes.coord(22)
        >>> coord
        (1.0, 0.5, 0.0, 0.0, 0.0, 0.0)

        Return all the node indices, coordinates, and angles of all the nodes.

        >>> nodes = mapdl.db.nodes
        >>> ind, coords, angles = nodes.all_asarray()
        >>> ind
        array([     1,      2,      3, ..., 270639, 270640, 270641], dtype=int32)

        >>> coords
        array([[0.    , 1.    , 0.    ],
               [0.    , 0.    , 0.    ],
               [0.    , 0.9875, 0.    ],
               ...,
               [0.9875, 0.975 , 0.925 ],
               [0.9875, 0.975 , 0.95  ],
               [0.9875, 0.975 , 0.975 ]])

        >>> angles
        array([[0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.],
               ...,
               [0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.]])

        """
        if self._nodes is None:
            from .nodes import DbNodes  # here to avoid circular import

            self._nodes = DbNodes(self)
        return self._nodes

    @property
    @check_mapdl_db_is_alive
    def elems(self):
        """
        MAPDL database element interface.

        Returns
        -------
        :class:`DbElems <ansys.mapdl.core.database.elems.DbElems>`

        Examples
        --------
        Create a MAPDL database element instance.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> elems = mapdl.db.elems
        >>> print(elems)
        MAPDL Database Elements
            Number of elements:          64
            Number of selected elements: 64
            Maximum element number:      64

        Return the element information of element 1.

        >>> elems = mapdl.db.elems
        >>> elem_info = elems.get(1)

        Return the nodes belonging to the element.

        >>> elem_info.nodes
        [2, 27, 37, 8]

        Return the element data.

        >>> elem_info.elmdat
        [1, 1, 1, 1, 0, 0, 14, 0, 0, 0]

        """
        if self._elems is None:
            from .elems import DbElems  # here to avoid circular import

            self._elems = DbElems(self)
        return self._elems



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import click

from ansys.mapdl.core.cli import main


@main.command(
    short_help="List MAPDL running instances.",
    help="""This command list MAPDL instances""",
)
@click.option(
    "--instances",
    "-i",
    is_flag=True,
    flag_value=True,
    type=bool,
    default=False,
    help="Print only instances",
)
@click.option(
    "--long",
    "-l",
    is_flag=True,
    flag_value=True,
    type=bool,
    default=False,
    help="Print all info.",
)
@click.option(
    "--cmd",
    "-c",
    is_flag=True,
    flag_value=True,
    type=bool,
    default=False,
    help="Print cmd",
)
@click.option(
    "--location",
    "-cwd",
    is_flag=True,
    flag_value=True,
    type=bool,
    default=False,
    help="Print running location info.",
)
def list_instances(instances, long, cmd, location):
    import psutil
    from tabulate import tabulate

    # Assuming all ansys processes have -grpc flag
    mapdl_instances = []

    def is_valid_process(proc):
        valid_status = proc.status() in [
            psutil.STATUS_RUNNING,
            psutil.STATUS_IDLE,
            psutil.STATUS_SLEEPING,
        ]
        valid_ansys_process = ("ansys" in proc.name().lower()) or (
            "mapdl" in proc.name().lower()
        )
        # Early exit to avoid checking 'cmdline' of a protected process (raises psutil.AccessDenied)
        if not valid_ansys_process:
            return False

        grpc_is_active = "-grpc" in proc.cmdline()
        return valid_status and valid_ansys_process and grpc_is_active

    for proc in psutil.process_iter():
        # Check if the process is running and not suspended
        try:
            if is_valid_process(proc):
                # Checking the number of children we infer if the process is the main process,
                # or one of the main process thread.
                if len(proc.children(recursive=True)) < 2:
                    proc.ansys_instance = False
                else:
                    proc.ansys_instance = True
                mapdl_instances.append(proc)

        except (psutil.NoSuchProcess, psutil.ZombieProcess) as e:
            continue

    # printing
    table = []

    if long:
        cmd = True
        location = True

    if instances:
        headers = ["Name", "Status", "gRPC port", "PID"]
    else:
        headers = ["Name", "Is Instance", "Status", "gRPC port", "PID"]

    if cmd:
        headers.append("Command line")
    if location:
        headers.append("Working directory")

    def get_port(proc):
        cmdline = proc.cmdline()
        ind_grpc = cmdline.index("-port")
        return cmdline[ind_grpc + 1]

    table = []
    for each_p in mapdl_instances:
        if instances and not each_p.ansys_instance:
            continue

        proc_line = []
        proc_line.append(each_p.name())

        if not instances:
            proc_line.append(each_p.ansys_instance)

        proc_line.extend([each_p.status(), get_port(each_p), each_p.pid])

        if cmd:
            proc_line.append(" ".join(each_p.cmdline()))

        if location:
            proc_line.append(each_p.cwd())

        table.append(proc_line)

    print(tabulate(table, headers))



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Superelements:
    def se(self, file="", toler="", **kwargs):
        """Defines a superelement.

        APDL Command: SE

        Parameters
        ----------
        file
            The name (case sensitive) of the file containing the
            original superelement matrix created by the generation
            pass (Sename.SUB).  The default is the current Jobname.

        toler
            Tolerance used to determine if use pass nodes are
            noncoincident with master nodes having the same node
            numbers.  Defaults to 0.0001.  Use pass nodes will always
            be replaced by master nodes of the same node number.
            However, if a use pass node is more than TOLER away from
            the corresponding master node, a warning is generated.

        Notes
        -----
        Defines a superelement by reading in the superelement matrices
        and master nodes from the superelement matrix file.  The
        matrix file (File.SUB) must be available from the substructure
        generation pass.  The proper element type (MATRIX50) must be
        active [TYPE] for this command.  A scratch file called
        File.SORD showing the superelement names and their
        corresponding element numbers is also written.
        """
        return self.run(f"SE,{file},,,{toler}", **kwargs)

    def sedlist(self, sename="", kopt="", **kwargs):
        """Lists the DOF solution of a superelement after the use pass.

        APDL Command: SEDLIST

        Parameters
        ----------
        sename
            Name of the superelement in Jobname.DSUB to be listed.  If a
            number, it is the element number of the superelement as used in the
            use pass.  If ALL, list results for all superelements.

        kopt
            List key:

            0 - List summary data only.

            1 - List full contents.  Be aware that the listing may be extensive.

        Notes
        -----
        Lists the degree of freedom solution of a superelement after the
        substructure use pass.  Results may be listed for any superelement on
        File.DSUB.

        This command is valid in any processor.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"SEDLIST,{sename},{kopt}"
        return self.run(command, **kwargs)

    def selist(self, sename="", kopt="", kint="", **kwargs):
        """Lists the contents of a superelement matrix file.

        APDL Command: SELIST

        Parameters
        ----------
        sename
            The name (case-sensitive) of the superelement matrix file created
            by the substructure generation pass (Sename.SUB).   Defaults to the
            current Jobname.  If a number, it is the element number of the
            superelement as used in the use pass.

        kopt
            List key:

            0 - List summary data only.

            1 - List contents, except load vectors and matrices.

            2 - List contents, except matrices.

            3 - List full contents.  Be aware that the listing may be extensive.

        kint
            Integer printout format key:

            OFF - Default.

            ON - Long format for large integers.

        Notes
        -----
        This command is valid in any processor.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"SELIST,{sename},{kopt},{kint}"
        return self.run(command, **kwargs)

    def sesymm(self, sename="", ncomp="", inc="", file="", ext="", **kwargs):
        """Performs a symmetry operation on a superelement within the use pass.

        APDL Command: SESYMM

        Parameters
        ----------
        sename
            The name (case-sensitive) of the superelement matrix file
            created by the substructure generation pass (Sename.SUB).
            Defaults to the current Jobname.  If a number, it is the
            element number of a previously defined superelement in the
            current use pass.

        ncomp
            Symmetry key:

            X - X symmetry (default).

            Y - Y symmetry.

            Z - Z symmetry.

        inc
            Increment all nodes in the superelement by INC.

        file
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).
            An unspecified directory path defaults to the working
            directory; in this case, you can use all 248 characters
            for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        Performs a symmetry operation on a superelement within the
        substructure use pass by reversing the sign of component Ncomp
        in the global Cartesian coordinate system.  The node numbers
        are incremented by INC.  The new superelement is written to
        File.SUB in the current directory (by default).  All master
        node nodal coordinate systems must be global Cartesian (no
        rotated nodes allowed).

        The maximum number of transformations for a given superelement
        is five (including SETRAN, SESYMM, and the large rotation
        transformation if NLGEOM is ON in the use pass).

        Distributed ANSYS Restriction: This command is not supported
        in Distributed ANSYS.
        """
        command = f"SESYMM,{sename},{ncomp},{inc},{file},{ext}"
        return self.run(command, **kwargs)

    def setran(
        self,
        sename="",
        kcnto="",
        inc="",
        file="",
        ext="",
        dx="",
        dy="",
        dz="",
        norot="",
        **kwargs,
    ):
        """Creates a superelement from an existing superelement.

        APDL Command: SETRAN

        Parameters
        ----------
        sename
            The name (case-sensitive) of the file containing the original
            superelement matrix created by the generation pass (Sename.SUB).
            The default is the current Jobname.  If Sename is a number, it is
            the element number of a previously defined superelement in the
            current use pass.

        kcnto
            The reference number of the coordinate system to where the
            superelement is to be transferred. The default is the global
            Cartesian system.  Transfer occurs from the active coordinate
            system.

        inc
            The node offset.  The default is zero.  All new element node
            numbers are offset from those on the original by INC.

        file
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        dx, dy, dz
            Node location increments in the global Cartesian coordinate system.
            Defaults to zero.

        norot
            Node rotation key:

            0 - The nodal coordinate systems of the transferred superelement rotate into the
                KCNTO system. (That is, the nodal coordinate systems rotate
                with the superelement.) The superelement matrices remain
                unchanged. This value is the default.

            1 - The nodal coordinate systems do not rotate. (That is, they remain fixed in
                their original global orientation.)  The superelement matrices
                and load vectors are modified if any rotations occur.

        Notes
        -----
        The SETRAN command creates a superelement from an existing superelement
        and writes the new element to a file.  You can then issue an SE command
        to read the new element (during the use pass).

        You can create a superelement from an original by:

        Transferring the original's geometry from the active coordinate system
        into another coordinate system (KCNTO)

        Offsetting its geometry in the global Cartesian coordinate system (DX,
        DY, and DZ )

        Offsetting its node numbers (INC).

        A combination of methods is valid. If you specify both the geometry
        transfer and the geometry offset, the transfer occurs first.

        If you specify rotation of the transferred superelement's nodal
        coordinate systems into the KCNTO system (NOROT = 0), the rotated nodes
        cannot be coupled via the CP command; in this case, issue the CE
        command instead. If you specify no rotation of the nodal coordinate
        systems (NOROT = 1) for models with displacement degrees of freedom,
        and KCNTO is not the active system, the superelement Sename must have
        six MDOF at each node that has MDOF; therefore, only elements with all
        six structural DOFs are valid in such cases.

        There is no limit to the number of copies that can be made of a
        superelement, provided the copies are all generated from the same
        original superelement. However, nested copies are limited to five. In
        other words, the total number of different Sename usages on the SETRAN
        and SESYMM commands is limited to five.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"SETRAN,{sename},{kcnto},{inc},{file},{ext},,{dx},{dy},{dz},{norot}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class SpecialPurpose:
    def aerocoeff(
        self,
        aeromodetype="",
        aeromappedfilenames="",
        aerospecs="",
        aeroscalar="",
        nblades="",
        autofileread="",
        **kwargs,
    ):
        """Computes the aero-damping and stiffness coefficients and
        writes them to an APDL array.

        APDL Command: AEROCOEFF

        Parameters
        ----------
        aeromodetype
            Mode type to be used.

            * ``"blade"`` : Non-cyclic cantilevered blade mode (default)

        aeromappedfiles
            Name of string array containing file names of mapped pressures
            from CFD. The file names should be ordered to correspond to
            the AeroSpecs array.

        aerospecs
            Name of numerical array containing data organized to correspond to
            the AeroMappedFiles array.

        aeroscalar
            Scaling value(s) to handle any modal scaling difference
            between structural and CFD modes. The values can be entered as
            a scalar or 1-dimensional array. (each scaling value defaults
            to 1)

        nblades
            Number of blades.

        autofileread
            Key to automatically read and use values from CFD file header.

            *  0 (OFF or NO) : Do not read scaling values or nodal diameter
               from the CFD file header. (default)
            * 1 (ON or YES) : Read scaling values (labeled Mode Multiplier
              in CFD file) from CFD file header. The scaling values read
              will be used in calculations and the AeroScalar input will
              be ignored. The nodal diameter values will be used to cross
              check the value of i (input through AeroSpecs array).

        Notes
        -----
        The AEROCOEFF command is designed to generate an array of
        aerodynamic coefficients that can be used in a cyclic
        mode-superposition harmonic response analysis using the CYCFREQ
        , AERO command to represent aerodynamic stiffness and
        damping. These aerodynamic coefficients can also be used in a damped
        modal analysis phase (CYCFREQ, MODAL) of a cyclic
        mode-superposition harmonic solve. An APDL array called
        JobnameAeroArray is generated using the AEROCOEFF command.  This array
        is compatible with the array needed for the CYCFREQ, AERO command.
        The format of the written array follows that of the CYCFREQ, AERO
        command. The array is formatted as follows:
        ``i, m, n, V_real, V_imag``

        where
        * ``i`` = the i th interblade phase angle (IBPA)
        * ``m`` = the m th vibrating blade mode
        * ``n`` = the n th blade mode generating the pressure oscillations

        ``Vreal`` and ``V_imag`` = the real and imaginary coefficients.

        Prior to issuing the AEROCOEFF command, a non-cyclic cantilevered
        blade modal analysis must be run, either stress-free or
        prestressed using linear perturbation. For more information, see
        Modal Analysis in the Structural Analysis Guide. The file
        requirements for the AEROCOEFF command are the same as those
        needed for modal restart as described in Modal Analysis Restart.
        The AeroSpecs values are specified in a 3xr array ``*DIM``,
        where r is a positive integer equal to the number of interblade
        phase angles and the pressure modes solved for in the CFD
        analysis.

        Each row has the structure: ``i, m, n``
        where
        * ``i`` = the i th interblade phase angle (IBPA)
        * ``m`` = the m th vibrating blade mode
        * ``n`` = the n th blade mode generating the pressure oscillations

        At least one aerodynamic damping coefficient must be specified for
        each IBPA (equal to the number of blades) while keeping and
        constant. If a value is not specified, the program writes an array
        value of zero for both and . The values of and are relative to the
        modes computed in the re- quired modal analysis.

        The number of AeroScalar values must be equal to the number of
        pressure modes ( from Aero- Specs). If the number of AeroScalar
        values is greater than 1, the values must be entered by defining
        an array ``*DIM`` and entering the array name in the AeroScalar
        field. For a discussion of how AeroScalar values are computed, see
        Scaling Aerodynamic Coupling Coefficients.

        The value for nBlades should be equal to the number of sectors of
        the system. If there are multiple blades per cyclic sector, then
        the combination of blades on the single sector will have an aero
        coefficient value. In this case, each blade will not have a
        distinct aero coefficient.
        """
        command = f"AEROCOEFF,{aeromodetype},{aeromappedfilenames},{aerospecs},{aeroscalar},{nblades},{autofileread}"
        return self.run(command, **kwargs)

    def cint(
        self,
        action="",
        par1="",
        par2="",
        par3="",
        par4="",
        par5="",
        par6="",
        par7="",
        **kwargs,
    ):
        """Defines parameters associated with fracture parameter calculations

        APDL Command: CINT

        Parameters
        ----------
        action
            Specifies action for defining or manipulating initial crack data:

            NEW
                Initiate a new calculation and assign an ID.

            CTNC
                Define the crack tip node component.

            CENC
                Define the crack extension node component, the crack tip node,
                and the crack extension direction.

            TYPE
                Define the type of calculation to perform.

            DELE
                Delete the ``CINT`` object associated with the specified ID.

            NCON
                Define the number of contours to be calculated in the
                contour-integral calculation.

            SYMM
                Indicate whether the crack is on a symmetrical line or plane.

            NORM
                Define the crack plane normal.

            UMM
                Activate or deactivate the unstructured mesh method.

            EDIR
                Crack-assist extension direction.

            PLOT
                Plots crack front and crack tip coordinate system.

            CXFE
                Define the crack tip element or crack front element set.
                Valid for XFEM-based crack-growth analysis only.

            RADIUS
                Define the radius at which the given value is to be evaluated.
                Valid for XFEM-based crack-growth analysis only.

            RSWEEP
                Define the minimum and maximum sweep angle from existing crack
                direction. Valid for XFEM-based crack-growth analysis only.
        """
        command = f"CINT,{action},{par1},{par2},{par3},{par4},{par5},{par6},{par7}"
        return self.run(command, **kwargs)

    def cycexpand(self, wn="", option="", value1="", value2=""):
        """Graphically expands displacements, stresses and strains of a
        cyclically symmetric model.

        APDL Command: /CYCEXPAND

        Parameters
        ----------
        wn
            The window number to which the expansion applies. Valid values are
            1 through 5. The default value is 1. The window number applies
            only to the ``AMOUNT`` argument.

        option
            One of the following options:

            ON
                Activates cyclic expansion using the previous settings (if any).
                If no previous settings exist, this option activates the default
                settings. This option is default.

            DEFAULT
                Resets cyclic expansion to the default settings.

            OFF
                Deactivates cyclic expansion.

            STATUS
                Lists the current cyclic expansion settings.

            AMOUNT
                The number of repetitions or the total angle.

                Value1
                    ``NREPEAT``

                Value2
                    The number of repetitions. The default is the total number
                    of sectors in 360 degrees.

                or

                Value1
                    ANGLE

                Value2
                    The total angle in degrees. The default is 360.

            WHAT
                A specified portion or subset of the model to expand:

                Value1
                    The component name of the elements to expand. The default
                    is all selected components.

            EDGE
                Sector edge display key. Possible ``Value1`` settings are:

                -1
                    Suppresses display of edges between sectors even if the
                    cyclic count varies between active windows. This setting is
                    not valid for cyclic mode-superposition (``MSUP``) harmonic
                    analyses.

                    .. warning::
                        Plots with fewer than the maximum number of
                        repetitions may have missing element faces at the sector
                        boundaries.

                0 or OFF
                    Averages stresses or strains across sector boundaries. This
                    value is the default (although the default reverts to 1 or
                    ON if the cyclic count varies between active windows).

                1 or ON
                    No averaging of stresses or strains occurs, and sector
                    boundaries are shown on the plot. This setting is not valid
                    for cyclic ``MSUP`` harmonic analyses.

            PHASEANG
                Possible ``Value1`` settings are:

                n
                    The phase angle shift in degrees. The valid range for n is 0
                    through 360. The default is 0. For a modal solution, this
                    value is typically the phase angle obtained via the
                    ``CYCPHASE`` command. The expanded modal results are printed
                    or displayed for the specified phase angle shift.

                AMPLITUDE (or n ≥ 360)
                    The amplitude is reported, except for the following
                    circumstances where the amplitude solution is not valid:

                    * non-component results (such as equivalent stress)
                    * modal analyses (no amplitude is calculated, and the
                      expanded modal results are printed or displayed at a
                      phase angle of 360º).

                SWEEP
                    For a mode-superposition harmonic solution, the maximum
                    values across a phase angle sweep are reported.

        Notes
        -----
        In preprocessing, the ``/CYCEXPAND`` command verifies a cyclically
        symmetric model by graphically expanding it partially or through the
        full 360 degrees.

        For the postprocessing plot nodal solution (``PLNSOL``) operation, the
        command graphically expands displacements, stresses and strains of a
        cyclically symmetric model partially or though the full 360 degrees by
        combining the real (original nodes and elements) and imaginary
        (duplicate nodes and elements) parts of the solution.

        For the print nodal solution (``PRNSOL``) operation, the command expands
        the printed output of displacements or stresses on a sector-by-sector
        basis. To learn more about specific ``PRNSOL`` behaviors in cyclic
        analyses, see Using the ``/CYCEXPAND`` Command in the *Cyclic Symmetry
        Analysis Guide*.

        Use of the ``/CYCEXPAND`` command does not change the database. The
        command does not modify the geometry, nodal displacements or element
        stresses.

        The command affects element and result plots only. It has no effect on
        operations other than plot element solution (``PLESOL``), plot nodal
        solution (``PLNSOL``), print nodal solution (`PRNSOL`), and calculate
        harmonic solution (``CYCCALC``). Operations other than ``PLESOL``,
        ``PLNSOL``, ``PRNSOL``, or ``CYCCALC`` work on the unprocessed real and
        imaginary parts of a cyclic symmetry solution

        If you issue a ``/CYCEXPAND,,OFF`` command, you cannot then expand the
        model by simply issuing another ``/CYCEXPAND`` command (for example, to
        specify an ``NREPEAT`` value for the number of repetitions). In such a
        case, you must specify ``/CYCEXPAND,,ON``, which activates expansion
        using the previous settings (if any) or the default settings.

        The command requires PowerGraphics and will turn PowerGraphics on
        (``/GRAPHICS,POWER``) if not already active. Any setting which bypasses
        PowerGraphics (for example, ``/PBF``) also bypasses cyclic expansion; in
        such cases, the ``/CYCEXPAND`` command displays unprocessed real and
        imaginary results.

        The ``CYCPHASE`` command uses full model graphics (``/GRAPHICS,FULL``)
        to compute peak values. Because of this, there may be slight differences
        between max/min values obtained with ``CYCPHASE``, and those obtained
        via ``/CYCEXPAND``, which uses power graphics (``/GRAPHICS,POWER``).

        For ``PHASEANG = AMPLITUDE`` (or 360) with a cyclic full harmonic
        solution, the only appropriate coordinate system is the solution
        coordinate system (``RSYS,SOLU``).

        Load case operations (``LCOPER``) are not valid during a cyclic
        expansion using ``/CYCEXPAND``.

        To learn more about analyzing a cyclically symmetric structure, see the
        *Cyclic Symmetry Analysis Guide*.
        """
        command = f"/CYCEXPAND, {wn}, {option}, {value1}, {value2}"
        return self.run(command, **kwargs)

    def cycfreq(
        self,
        option="",
        value1="",
        value2="",
        value3="",
        value4="",
        value5="",
        **kwargs,
    ):
        """Specifies solution options for a cyclic symmetry mode-superposition

        APDL Command: CYCFREQ
        harmonic analysis.

        Parameters
        ----------
        option
            One of the following options:

            AERO
                Specify the array containing the aerodynamic damping
                coefficients.

                Value1
                    The name of the array containing the aerodynamic stiffness
                    damping coefficients.

            BLADE
                Blade information required for a mistuning analysis.

                Value1
                    The name of the nodal component containing the blade
                    boundary nodes at the
                    blade-to-disk interface. Also include boundary nodes at
                    any shroud interfaces.

                Value2
                    The name of the element component containing the blade
                    elements.

                Value3
                    The number of blade modes to include in the CMS reduction.

                Value4
                    The lower bound of the frequency range of interest. This
                    value is optional.

                Value5
                    The upper bound of the frequency range of interest. This
                    value is optional.

            DEFAULT
                Set the default cyclic harmonic solution settings.

            EO
                Excitation engine order.

                Value1
                    The value of the excitation order, which must be an integer.
                    The loadings on the other sectors will be related to the
                    loading on the basic sector based on the engine order phase
                    shift.

                Value2
                    The name of the Mechanical APDL array containing the modal
                    forces corresponding to the modes kept in the
                    mode-superpostion analysis.

            MIST
                Mistuning parameters.

                Value1
                    The type of mistuning:

                    K
                        Stiffness (frequency) mistuning

                Value2
                    The name of the array containing the stiffness mistuning
                    parameters.

            MODAL
                Specifies if a damped modal analysis should be performed on the
                reduced system.

                Value1
                    On/Off key.

                    0 (OFF or NO)
                        No modal solution. Perform the harmonic solution.

                    1 (ON or YES)
                        Perform a damped modal analysis of the reduced system in
                        order to obtain the complex frequencies. The harmonic
                        solution is not performed.

                Value2
                    Number of modes for the damped modal analysis.

                Value3
                    The beginning, or lower end, of the frequency range of
                    interest (in Hz).

                Value4
                    The ending, or upper end, of the frequency range of interest
                    (in Hz).

            RESTART
                Defines the point at which to restart the harmonic analysis.

                Value1
                    The restart point:

                    OFF
                        No restart (default)

                    SWEEP
                        Restart for a new frequency sweep range (``HARFRQ``)

                    MIST
                        Restart for new mistuning parameters (new mistuning
                        arrays)

            USER
                Causes the program to call for a user-defined solution.

                Value1-5
                    Values passed down to the user-defined solution.

            STATUS
                List the harmonic solution option settings active for the cyclic
                model.

        Notes
        -----
        The program solves a cyclically symmetric model (set up via the ``CYCLIC``
        command during preprocessing) at the harmonic indices specified via the
        ``CYCOPT`` command.

        The aerodynamic coefficients are specified in a 5x(Nxr) array (``*DIM``),
        where N is the number of blades and r can be any positive integer. Each
        column has the structure:

        where:

        One aerodynamic damping coefficient must be specified for each ``IBPA``
        (equal to the number of blades) while keeping m and n constant.

        For constant (frequency-independent) mistuning, the stiffness
        parameters are specified in an Nx1 array (``*DIM``) where N is the number
        of blades.

        For stiffness mistuning, each row entry represents the deviation of
        Young's modulus from nominal,  (or equivalently, the ratio of the
        frequency deviation squared). Each frequency can also be independently
        mistuned, in which case the array is NxM, where M is the number of
        blade frequencies (Value3 of ``CYCFREQ,BLADE``). The entries in each row
        therefore correspond to the ratio of the mistuned frequency to the
        tuned frequency squared minus one:

        The ``USER`` option activates the solution macro ``CYCMSUPUSERSOLVE.MAC``.
        The normal solution is skipped. You may implement your own mistuning
        solution using APDL and APDL Math operations, or call your own program
        for the solution.

        The ``CYCFREQ`` command is valid in the preprocessing and solution stages
        of an analysis.

        The ``CYCFREQ,MODAL,ON`` command writes modal frequencies to the output
        file. No other postprocessing is available for this modal solve.

        When using ``CYCFREQ,RESTART``, only mistuning parameters or frequency
        range may be changed. All other changes in parameters are ignored.

        To learn more about analyzing a cyclically symmetric structure, see the
        Cyclic Symmetry Analysis Guide.
        """
        command = f"CYCFREQ,{option},{value1},{value2},{value3},{value4},{value5}"
        return self.run(command, **kwargs)

    def cyclic(
        self,
        nsector="",
        angle="",
        kcn="",
        name="",
        usrcomp="",
        usrnmap="",
        **kwargs,
    ):
        """Specifies a cyclic symmetry analysis.

        APDL Command: CYCLIC

        Parameters
        ----------
        nsector
            The number of sectors in the full 360 degrees, or one of the
            following options:

            STATUS
                Indicates the current cyclic status.

            OFF
                Resets model to normal (non-cyclic) status and removes the
                duplicate sector if it exists. This option also deletes
                automatically detected edge components (generated when ``USRCOMP = 0``).

            UNDOUBLE
                Removes the duplicate sector if it exists. The duplicate sector
                is created during the solution (SOLVE) stage of a modal cyclic
                symmetry analysis.

        angle
            The sector angle in degrees.

        kcn
            An arbitrary reference number assigned to the cyclic coordinate
            system.  The default value of 0 specifies automatic detection.

        name
            The root name of sector low- and high-edge components (line, area,
            or node components). The default root name (when ``USRCOMP`` = 0) is
            "CYCLIC". A root name that you specify can contain up to 11
            characters.

        usrcomp
            The number of pairs of user-defined low- and high-edge components
            on the cyclic sector (if any). The default value of 0 specifies
            automatic detection of sector edges; however, the automatic setting
            is not valid in all cases. (For more information, see the Notes
            section below.) If the value is greater than 0, no verification of
            user-defined components occurs.

        usrnmap
            The name of a user-defined array specifying the matching node pairs
            between the sector low and high edges. Valid only when ``USRCOMP`` = 0.
            Skips the automatic detection of sector edges. Node pairs may be
            input in any order, but the low edge node must be the first entry
            in each pair.

        Notes
        -----
        You can input your own value for ``NSECTOR``, ``ANGLE`` or ``KCN``; if
        you do so, the command verifies argument values before executing.

        When ``USRCOMP = 0`` and ``USRNMAP = blank`` (default), the ``CYCLIC`` command
        automatically detects low- and high-edge components for models
        comprised of any combination of line, area, or volume elements. If a
        solid model exists, however, the  command uses only the lines, areas,
        and/or volumes to determine the  low- and high-edge components; the
        elements, if any, are ignored.

        Nodes will be automatically rotated unless ``CYCOPT,USERROT,YES`` has been
        specified.

        If you issue a ``CYCOPT,TOLER`` command to set a tolerance for edge-
        component pairing before issuing the ``CYCLIC`` command, the ``CYCLIC`` command
        uses the specified tolerance when performing automatic edge-component
        detection.

        For 2-D models, autodetection does not consider the ``CSYS,5`` or ``CSYS,6``
        coordinate system specification.  Autodetection for 180 degree (two-
        sector) models is not possible unless a central hole exists.

        The ``CYCLIC`` command sets values and keys so that, if possible, the area-
        mesh (``AMESH``) or volume-mesh (``VMESH``) command meshes the sector with
        matching node and element face patterns on the low and high edges. (The
        command has no effect on any other element-creation command.)

        Issue the ``CYCLIC`` command prior to the meshing command to, if possible,
        produce a mesh with identical node and element patterns on the low and
        high sector edges. Only the ``AMESH`` or ``VMESH`` commands can perform
        automated matching. (Other meshing operation commands such as VSWEEP
        cannot.) If you employ a meshing operation other than ``AMESH`` or ``VMESH``,
        you should ensure that node and element face patterns match, if
        desired. The ``CYCLIC`` command output indicates whether each edge-
        component pair has or can produce a matching node pair.

        A cyclic solution (via the ``SOLVE`` command) allows dissimilar mesh
        patterns on the extreme boundaries of a cyclically symmetric model. The
        allowance for dissimilar patterns is useful when you have only finite-
        element meshes for your model but not the geometry data necessary to
        remesh it to obtain identical node patterns. In such cases, it is
        possible to obtain solution results, although perhaps at the expense of
        accuracy. A warning message appears because results may be degraded
        near the sector edges.

        The constraint equations (CEs) that tie together the low and high edges
        of your model are generated at the solution stage of the analysis from
        the low- and high-edge components (and nowhere else). You should verify
        that automatically detected components are in the correct  locations
        and that you can account for all components; to do so, you can list
        (``CMLIST``) or plot (``CMPLOT``) the components.

        If you issue the ``CYCLIC`` command after meshing and have defined element
        types with rotational degrees of freedom (DOFs), ANSYS generates cyclic
        CEs for rotational DOFs that may not exist on the sector boundaries.
        Issue the ``CYCOPT,DOF`` command to prevent unused rotational terms from
        being generated.

        Modal cyclic symmetry analysis is supported by the following
        eigensolvers:

        * Block Lanczos (``MODOPT,LANB``)
        * PCG Lanczos (``MODOPT,LANPCG``)
        * Super Node (``MODOPT,SNODE``)
        * Subspace (``MODOPT,SUBSP``)

        To learn more about analyzing a cyclically symmetric structure, see the
        Cyclic Symmetry Analysis Guide.

        When using the: ``CYCLIC``: command to automatically detect the sector, if
        an area is defined with the: AL: command, the lines need to be oriented
        to form the closed curve.
        """
        command = f"CYCLIC,{nsector},{angle},{kcn},{name},{usrcomp},{usrnmap}"
        return self.run(command, **kwargs)

    def cycopt(
        self,
        option="",
        value1="",
        value2="",
        value3="",
        value4="",
        value5="",
        value6="",
        value7="",
        **kwargs,
    ):
        """Specifies solution options for a cyclic symmetry analysis.

        APDL Command: CYCOPT

        Parameters
        ----------
        option
            One of the following options:

            BCMULT - Controls whether cyclic sector array parameter names are reused or created new
                     for multiple entities.

            Value1 - The flag value.

            0 (OFF or NO) - Create new array parameter names (default)

            1(ON or YES) - Reuse array parameter names

            COMBINE - For linear static cyclic symmetry analysis with non-cyclically symmetric
                      loading only, expands and combines all harmonic index
                      solutions and writes them to the results file during the
                      solution phase of the analysis.

            Value1 - The flag value.

            0 (OFF or NO) - Disable combining of harmonic index solutions (default)

            1 (ON or YES) - Enable combining of harmonic index solutions

            DEFAULT - Set the default cyclic solution settings.

            DOF - The degrees of freedom to couple from the nodes on the low sector boundary to
                  nodes on the high boundary:

            Value1 - The component pair ID number.

            Value2, Value3, Value4,  . . . ,  Value7 - The constraint-equation/-coupling degree of freedom (DOF) for this pair. Repeat
                              the command to add other DOFs. The default is
                              constraint-equation/-coupling all applicable
                              DOFs.

            FACETOL - Tolerance for inclusion of surface nodes into your basic sector. Autodetect
                      defaults to 15°, accommodating most sections. Specify a
                      new Value1 only when extreme cut angles or complex model
                      geometry cause surface nodes to be excluded. See Notes
                      (below) for more information.

            ANSYS, Inc. recommends that successful auto-detection depends more on the value of ANGTOL than the value of FACETOL. Please refer to CYCOPT Auto Detection Tolerance Adjustments for Difficult Cases for more information about auto-detection and the CYCOPT command. - Value1

            The face tolerance applies only to auto detection from node/element models (already meshed and no solid model), and it defaults to 15°.   - HINDEX

            The harmonic index solution ranges for modal or buckling cyclic symmetry analyses. The SOLVE command initiates a cyclic symmetry solution sequence at the harmonic indices specified. (By default, the SOLVE command solves for all available harmonic indices.) Static and harmonic cyclic symmetry solutions always use all harmonic indices required for the applied loads. - EVEN / ODD

            For low-frequency electromagnetic analysis only, EVEN specifies a symmetric solution and ODD specifies an antisymmetric solution.  - The value you specify is based on the harmonic index: EVEN (default) indicates
                              harmonic index = 0, and ODD indicates harmonic
                              index = N / 2 (where N is an integer representing
                              the number of sectors in 360°). A value of ODD
                              applies only when N is an even number.

            The CYCOPT command with this HINDEX option is cumulative. To remove an option (for example, EVEN), issue this command: CYCOPT,HINDEX,EVEN,,,-1 - ALL

            Solve all applicable harmonic indices. - Note:  Value2 must be blank.

            Value1, Value2, Value3 - Solve harmonic indices in range Value1 through Value2 in steps of Value3.
                              Repeat the command to add other ranges. The
                              default solves all applicable harmonic indices.

            Value4 - The only valid value is -1. If specified, it removes Value1 through Value2 in
                     steps of Value3 from the set to solve. By default, if
                     Value4 = -1 then Value1 = 0, Value2 = 0, and Value3 = 1.

            Value5 - For static and harmonic analyses, the tolerance for determining if a Fourier
                     contribution of a load contributes to the response
                     (default = 1.0E-5).

            If Value5=STATIC, it forces the program to solve only the specified harmonic indices (even if a load may have a Fourier contribution in an index not specified). - LDSECT

            Restricts subsequently defined force loads and surface loads to a specified sector. The restriction remains in effect until you change or reset it. This option is not available for harmonic analyses based on mode-superposition (CYCOPT,MSUP,1) - Value1

            The sector number. A value other than 0 (default) is valid for a cyclic symmetry analysis with non-cyclically symmetric loading only. A value of 0 (or ALL) resets the default behavior for cyclic loading (where the loads are identical on all sectors). - MOVE

            Specifies if the program should move high- or low-edge component nodes paired within the specified tolerance (TOLER) to create precisely matching pairs.  - Value1

            The flag value. - 0

            Do not move edge component nodes (default) - 1 or HIGH

            Move the high-edge component nodes to precisely match the low-edge component nodes - -1 or LOW

            Move the low-edge component nodes to precisely match the high-edge component nodes - MSUP

            For modal cyclic symmetry analysis only, this flag is used to limit the results written to the Jobname.MODE and Jobname.RST files in preparation for a subsequent mode-superposition-based analysis. In a linear perturbation modal analysis, this option must be specified in the first load step of the preceding base analysis.  - m
        """
        command = f"CYCOPT,{option},{value1},{value2},{value3},{value4},{value5},{value6},{value7}"
        return self.run(command, **kwargs)

    def emsym(self, nsect="", **kwargs):
        """Specifies circular symmetry for electromagnetic sources.

        APDL Command: EMSYM

        Parameters
        ----------
        nsect
            The number of circular symmetry sections (defaults to 1).

        Notes
        -----
        Specifies the number of times to repeat electromagnetic sources for
        circular symmetry. Applies to SOURC36 elements and to coupled-field
        elements with electric current conduction results in the database.
        Sources are assumed to be equally spaced over 360° about the global
        Cartesian Z axis.

        This command is also valid in SOLUTION.
        """
        command = f"EMSYM,{nsect}"
        return self.run(command, **kwargs)

    def mstole(self, method="", namesurf="", namefluid="", **kwargs):
        """Adds two extra nodes from FLUID116 elements to SURF151 or SURF152

        APDL Command: MSTOLE
        elements for convection analyses.

        Parameters
        ----------
        method
            Mapping method:

            0 - Hybrid method (default).

            1 - Projection method.

            2 - Minimum centroid distance method.

        namesurf
            Component name for a group of SURF151 or SURF152 elements. The name
            must be enclosed in single quotes (e.g., 'COM152') when the MSTOLE
            command is manually typed in.

        namefluid
            Component name for a group of FLUID116 elements. The name must be
            enclosed in single quotes (e.g., 'COM116') when the MSTOLE command
            is manually typed in.

        Notes
        -----
        For convection analyses, the MSTOLE command adds two extra nodes from
        FLUID116 elements to SURF151 or SURF152 elements by employing the
        specified mapping method. In the hybrid method, the projection method
        is tried first and if it fails the centroid distance method is used.
        The SURF151 or SURF152 elements and the FLUID116 elements must be
        grouped into components and named using the CM command.

        The SURF151 or SURF152 extra node option must be set for two extra
        nodes (KEYOPT(5) = 2).

        For more information, see Using the Surface Effect Elements in the
        Thermal Analysis Guide.
        """
        command = f"MSTOLE,{method},{namesurf},{namefluid}"
        return self.run(command, **kwargs)

    def perbc2d(
        self,
        loc1="",
        loc2="",
        loctol="",
        r1="",
        r2="",
        tolr="",
        opt="",
        plnopt="",
        **kwargs,
    ):
        """Generates periodic constraints for 2-D planar magnetic field analyses.

        APDL Command: PERBC2D

        Parameters
        ----------
        loc1
            Constant coordinate location of the first plane of nodes.  For
            PLNOPT = 1 or 2, the constant coordinate location is the global
            Cartesian coordinate system [CSYS,0] location in the X or Y
            direction respectively.  For PLNOPT = 0, the location is the angle
            in the global cylindrical coordinate system [CSYS,1].

        loc2
            Constant coordinate location of the second plane of nodes.  For
            PLNOPT  = 1 or 2, the constant coordinate location is the global
            Cartesian coordinate system [CSYS,0] location in the X or Y
            direction respectively.  For PLNOPT = 0, the location is the angle
            (in degrees) in the global cylindrical coordinate system [CSYS,1].

        loctol
            Tolerance on the constant coordinate location for node selection.
            Defaults to .00001 for PLNOPT  = 1 or 2 and .001 degrees for PLNOPT
            = 0.

        r1
            Minimum coordinate location along the second plane of nodes.  For
            PLNOPT = 1 or 2, the coordinate location is the global Cartesian
            coordinate system location in the Y or X direction respectively.
            For PLNOPT = 0, the coordinate location is the radial coordinate
            value in the global cylindrical coordinate system.   Periodic
            conditions are not applied to nodes at this location.

        r2
            Maximum coordinate location along the second plane of nodes.  For
            PLNOPT = 1 or 2, the coordinate location is the global Cartesian
            coordinate system location in the Y or X direction respectively.
            For PLNOPT = 0, the coordinate location is the radial coordinate
            value in the global cylindrical coordinate system. Periodic
            conditions are not applied to nodes at this location.

        tolr
            Tolerance dimension on node selection along the plane of nodes.
            Defaults to .00001.

        opt
            Periodic option:

            0 - Odd symmetry (default).  Apply constraint equations such that AZ(i) = -AZ(j).

            1 - Even symmetry.  Apply node coupling such that AZ(i) = AZ(j).

        plnopt
            Symmetry plane option:

            0 - Planes of constant angle in the global cylindrical coordinate system [CSYS,1].

            1 - Planes parallel to the global Cartesian X axis [CSYS,0].

            2 - Planes parallel to the global Cartesian Y axis [CSYS,0].

        Notes
        -----
        PERBC2D invokes an ANSYS macro which generates periodic boundary
        condition constraints for 2-D planar magnetic field analysis.  The
        macro is restricted to node pairs sharing common coordinate values
        along symmetry planes separated by a constant coordinate value.  Planes
        (or lines) must lie at either constant  angles (PLNOPT = 0), constant X
        values (PLNOPT = 1), or constant Y values (PLNOPT = 2).  PERBC2D
        applies constraint equations (OPT = 0, odd symmetry) or node coupling
        (OPT = 1, even symmetry) to each node pair sharing a common coordinate
        value along the symmetry planes.  By default, periodic conditions are
        not applied at the first and last node pairs on the symmetry planes
        unless the input location values, R1 and R2, are adjusted to be less
        than or greater than the actual node coordinate values.  Nodes are
        selected for application of the constraints using the NSEL command with
        tolerances on the constant coordinate location (LOCTOL) and the
        coordinate location along the plane (RTOL).

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"PERBC2D,{loc1},{loc2},{loctol},{r1},{r2},{tolr},{opt},{plnopt}"
        return self.run(command, **kwargs)

    def physics(self, option="", title="", fname="", ext="", **kwargs):
        """Writes, reads, or lists all element information

        APDL Command: PHYSICS

        Parameters
        ----------
        option
            Specifies what to do with element information:

            WRITE - Write all appropriate element types, key options,
                    real constants, material properties, solution
                    analysis options, load step options, constraint
                    equations, coupled nodes, defined components, and
                    GUI preference settings to the file specified with
                    the Fname and Ext arguments.

            READ - Deletes all solution information (material
                   properties, solution options, load step options,
                   constraint equations, coupled nodes, results, and
                   GUI preference settings) then reads all the
                   information listed above into the ANSYS database
                   from the location specified by the Fname and Ext
                   arguments.

            LIST - Lists currently defined physics files and their titles.

            DELETE - Deletes a specified physics file and its title
                     from the database.

            CLEAR - Deletes all material properties, solution options,
                    load step options, constraint equations, coupled
                    nodes, results, and GUI preference settings from
                    the database. Does NOT clear the active physics
                    file title from the database.

            STATUS - Displays information about all active elements
                     and settings.

        title
            A user-defined title that quickly identifies a set of
            physics settings.  For example, you might use "Fluid,"
            "Structural," or "Magnetic" as titles.  A title can
            contain up to 64 characters. It can be entered in lower or
            upper case. Lower case is internally converted to upper
            case within the program.

        fname
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).
            An unspecified directory path defaults to the working
            directory; in this case, you can use all 248 characters
            for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        Use the PHYSICS command when you are performing a multiphysics
        analysis that involves two different disciplines (for example,
        magnetic and structural analyses) and you cannot solve both
        analyses simultaneously.  Once you have set up physics
        environments for both analyses, you can use the PHYSICS,READ
        command to change between the defined physics environments.
        For more information about doing multiphysics analyses, see
        Sequential Coupled-Field Analysis in the Coupled-Field
        Analysis Guide.

        The PHYSICS command outputs all solution information,
        including analysis options, to the Jobname.PHn file described
        above.  Although it also outputs components, the ANSYS program
        does not list entities (nodes, elements, lines, etc.).

        PHYSICS,WRITE will overwrite existing physics files with the
        same title (even if the name is different). In other words, if
        the directory has a physics file with the same title as the
        active physics file title, but a different name, the
        PHYSICS,WRITE command will overwrite the existing physics file
        and use the existing filename, not the filename specified on
        the PHYSICS,WRITE command.
        """
        command = f"PHYSICS,{option},{title},{fname},{ext}"
        return self.run(command, **kwargs)

    def race(self, xc="", yc="", rad="", tcur="", dy="", dz="", cname="", **kwargs):
        """Defines a "racetrack" current source.

        APDL Command: RACE

        Parameters
        ----------
        xc
            Location of the mid-thickness of the vertical leg along the working
            plane X-axis.

        yc
            Location of the mid-thickness of the horizontal leg along the
            working plane Y-axis.

        rad
            Radius of curvature of the mid-thickness of the curves in the
            racetrack source.  Defaults to .501 * DY

        tcur
            Total current, amp-turns (MKS), flowing in the source.

        dy
            In-plane thickness of the racetrack source.

        dz
            Out-of-plane thickness (depth) of the racetrack source.

        cname
            An alphanumeric name assigned as a component name to the group of
            SOURC36 elements created by the command macro.  Cname must be
            enclosed in single quotes in the RACE command line.  Cname may be
            up to 32 characters, beginning with a letter and containing only
            letters, numbers, and underscores.  Component names beginning with
            an underscore (e.g., _LOOP) are reserved for use by ANSYS and
            should be avoided.  If blank, no component name is assigned.

        Notes
        -----
        RACE invokes an ANSYS macro which defines a "racetrack" current source
        in the working plane coordinate system.  The current source is
        generated from bar and arc source primitives using the SOURC36 element
        (which is assigned the next available element type number).  The macro
        is valid for use in 3-D magnetic field analysis using a scalar
        potential formulation.  Current flows in a counterclockwise direction
        with respect  to the working plane.

        The diagram below shows you a racetrack current source.
        """
        return self.run(f"RACE,{xc},{yc},{rad},{tcur},{dy},{dz},,,{cname}", **kwargs)

    def sstate(
        self,
        action="",
        cm_name="",
        val1="",
        val2="",
        val3="",
        val4="",
        val5="",
        val6="",
        val7="",
        val8="",
        val9="",
        **kwargs,
    ):
        """Defines a steady-state analysis.

        APDL Command: SSTATE

        Parameters
        ----------
        action
            Action to perform for defining or manipulating steady-state
            analysis data:

            DEFINE - Define steady-state analysis data

            LIST - List current steady-state analysis data

            DELETE - Delete steady-state analysis data

        cm_name
            Element component name

        val1, ..., val9
            Input values (based on the Action type)

        Notes
        -----
        The SSTATE command specifies steady-state analysis parameters for the
        given element component. The program runs the steady-state analysis if
        the corresponding element key option is enabled for that element
        component.

        The command supports the following elements:

        SOLID185

        3-D 8-Node Structural Solid

        SOLID186

        3-D 20-Node Structural Solid

        SOLID187

        3-D 10-Node Tetrahedral Structural Solid

        SOLSH190

        3-D 8-Node Structural Solid Shell

        Degenerated shape (prism) option not supported.

        SOLID285

        3-D 4-Node Tetrahedral Structural Solid with Nodal Pressures

        For information about steady-state rolling for rebar and solid
        elements, see Steady State Rolling in the Mechanical APDL Theory
        Reference.

        The following data types can be defined:

        SPIN -- Steady-state spinning motion

        TRANSLATE -- Rigid body motion (velocity) that the spinning component
        is undergoing

        Define the steady-state spinning motion:

        SSTATE, DEFINE, CM_Name, SPIN, OMEGA, Method, Val4, Val5, Val6, Val7,
        Val8, Val9

        Spin velocity

        Method to use for defining the spin axis:

        Define the spin axis using two points:

        Val4, Val5, Val6 -- Coordinates of the first point

        Val7, Val8, Val9 -- Coordinates of the second point

        This definition method is currently the only option.

        This command defines a steady state spinning motion of 120 rad/s around
        the spin axis:

        In this case, two points with coordinates (0,0,0) and (0,1,0) define
        the spin axis in the global Y direction.

        Define the rigid body motion (velocity):

        SSTATE, DEFINE, CM_Name, TRANSLATE, Val2, Val3, Val4

        SSTATE, LIST, CM_Name

        Lists all steady-state analysis data defined on the specified element
        component. All data is listed if no component (CM_Name) is specified.

        SSTATE, DELETE, CM_Name

        Deletes all steady-state analysis data defined on the specified element
        component. All data is deleted if no component (CM_Name) is specified.
        """
        command = f"SSTATE,{action},{cm_name},{val1},{val2},{val3},{val4},{val5},{val6},{val7},{val8},{val9}"
        return self.run(command, **kwargs)

    def xfdata(
        self, enrichmentid="", lsm="", elemnum="", nodenum="", phi="", psi="", **kwargs
    ):
        """Defines a crack in the model by specifying nodal level set values

        APDL Command: XFDATA

        Parameters
        ----------
        enrichmentid
            Name of the enrichment specified via the associated XFENRICH
            command.

        lsm (or blank)
            Specify LSM to indicate that level set values (PHI and optional
            PSI) are specified.

        elemnum
            Element number.

        nodenum
            Node number associated with the specified element ELNUM.

        phi
            Signed normal distance of the node from the crack.

        psi
            Signed normal distance of the node from the crack tip (or crack front).
            Used only in the singularity- based XFEM method.

        Notes
        -----
        Issue the XFDATA command multiple times as needed to specify nodal
        level set values for all nodes of an element.

        This command is valid in PREP7 (/PREP7) only.
        """
        command = f"XFDATA,{enrichmentid},{lsm},{elemnum},{nodenum},{phi},{psi}"
        return self.run(command, **kwargs)

    def xfenrich(
        self,
        enrichmentid="",
        compname="",
        matid="",
        method="",
        radius="",
        snaptoler="",
        **kwargs,
    ):
        """Defines parameters associated with crack propagation using XFEM

        APDL Command: XFENRICH

        Parameters
        ----------
        enrichmentid
            An alphanumeric name assigned to identify the enrichment. The name
            can contain up to 32 characters and must begin with an alphabetic
            character. Alphabetic characters, numbers, and underscores are
            valid.

        compname
            Name of the element set component for which initial cracks are
            defined and possibly propagated.

        matid
            Material ID number referring to cohesive zone material behavior on
            the initial crack. If 0 or not specified, the initial crack is
            assumed to be free of cohesive zone behavior.

        method
            PHAN -- Use phantom-node-based XFEM (default).
            SING -- Use singularity-based XFEM.

        radius
            Radius defining the region around the crack tip encompassing the
            set of elements to be influenced by the crack-tip singularity effects.
            Default = 0.0. Used only in singularity-based XFEM.

        snaptoler
            Snap tolerance to snap the crack tip to the closest crack face along
            the extension direction. Default = 1.0E-6. Used only in singularity-based XFEM.

        Notes
        -----
        If MatID is specified, the cohesive zone behavior is described by the
        bilinear cohesive law.

        This command is valid in PREP7 (/PREP7) only.
        """
        command = f"XFENRICH,{enrichmentid},{compname},{matid}, {method}, {radius}, {snaptoler}"
        return self.run(command, **kwargs)

    def xflist(self, enrichmentid="", **kwargs):
        """Lists enrichment details and associated crack information

        APDL Command: XFLIST

        Parameters
        ----------
        enrichmentid or (blank)
            Name of the enrichment specified via the associated XFENRICH
            command. Specifying EnrichmentID is optional.

        Notes
        -----
        This command is valid in PREP7 (/PREP7) and SOLUTION (/SOLU).
        """
        command = f"XFLIST,{enrichmentid}"
        return self.run(command, **kwargs)



# Copyright (C) 2016 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

POINT_SIZE = 10

# Supported labels
BC_D = [
    "TEMP",
    "UX",
    "UY",
    "UZ",
    "VOLT",  # "MAG"
]
BC_F = [
    "HEAT",
    "FX",
    "FY",
    "FZ",
    "AMPS",
    "CHRG",
    # "FLUX",
    "CSGZ",
]  # TODO: Add moments MX, MY, MZ
FIELDS = {
    "MECHANICAL": ["UX", "UY", "UZ", "FX", "FY", "FZ"],
    "THERMAL": ["TEMP", "HEAT"],
    "ELECTRICAL": ["VOLT", "CHRG", "AMPS"],
}

FIELDS_ORDERED_LABELS = FIELDS["MECHANICAL"].copy()
FIELDS_ORDERED_LABELS.extend(FIELDS["THERMAL"])
FIELDS_ORDERED_LABELS.extend(FIELDS["ELECTRICAL"])

# All boundary conditions:
BCS = BC_D.copy()
BCS.extend(BC_F)

# Allowed entities to plot their boundary conditions
ALLOWED_TARGETS = ["NODES"]
